from inka2.helpers import parse_str_to_bool


def test_parse_str_to_bool():
    assert parse_str_to_bool("True") is True
    assert parse_str_to_bool("False") is False
    assert parse_str_to_bool("true") is True
    assert parse_str_to_bool("false") is False
    assert parse_str_to_bool("True ") is True
    assert parse_str_to_bool(" False") is False
    assert parse_str_to_bool("true ") is True
    assert parse_str_to_bool("false ") is False
    assert parse_str_to_bool(" True") is True
    assert parse_str_to_bool(" False") is False
    assert parse_str_to_bool(" true") is True
    assert parse_str_to_bool(" false") is False
    assert parse_str_to_bool("True") is True
    assert parse_str_to_bool("False") is False
    assert parse_str_to_bool("true") is True
    assert parse_str_to_bool("false") is False
    assert parse_str_to_bool("True ") is True
    assert parse_str_to_bool(" False") is False
    assert parse_str_to_bool("true ") is True
    assert parse_str_to_bool("false ") is False
    assert parse_str_to_bool(" True") is True
    assert parse_str_to_bool(" False") is False
    assert parse_str_to_bool(" true") is True
    assert parse_str_to_bool(" false") is False
    assert parse_str_to_bool("True") is True
    assert parse_str_to_bool("False") is False
    assert parse_str_to_bool("true") is True
    assert parse_str_to_bool("false") is False
    assert parse_str_to_bool("True ") is True
    assert parse_str_to_bool(" False") is False
    assert parse_str_to_bool("true ") is True
    assert parse_str_to_bool("false ") is False
    assert parse_str_to_bool(" True") is True
    assert parse_str_to_bool(" False") is False
    assert parse_str_to_bool(" true") is True
    assert parse_str_to_bool(" false") is False
    assert parse_str_to_bool("True") is True
    assert parse_str_to_bool("False") is False
    assert parse_str_to_bool("true") is True
    assert parse_str_to_bool("false") is False
    assert parse_str_to_bool("True ") is True



import configparser
import os
from pathlib import Path
from typing import List, Union

from rich.table import Column, Table


class Config:
    """Class for working with the configuration file."""

    _default_deck = "Default"
    _default_folder = ""
    _default_profile = ""
    _default_path = ""
    _default_basic_type = "Inka Basic"
    _default_front_field = "Front"
    _default_back_field = "Back"
    _default_cloze_type = "Inka Cloze"
    _default_cloze_field = "Text"
    _default_highlight_style = "monokai"
    _default_escape_html = False
    _add_filename = False

    def __init__(self, config_path: Union[str, Path]):
        self._config = configparser.ConfigParser()
        self._config_path = config_path

        if not os.path.exists(self._config_path):
            self._create_default()
        else:
            self._read()

    def reset(self):
        """Reset config file to default state"""
        self._config = configparser.ConfigParser()
        self._create_default()

    def _create_default(self):
        """Create default configuration file"""
        config_dict = {
            "defaults": {
                "profile": self._default_profile,
                "deck": self._default_deck,
                "folder": self._default_folder,
                "escape_html": self._default_escape_html,
                "add_filename": self._default_escape_html,
            },
            "anki": {
                "path": self._default_path,
                "basic_type": self._default_basic_type,
                "front_field": self._default_front_field,
                "back_field": self._default_back_field,
                "cloze_type": self._default_cloze_type,
                "cloze_field": self._default_cloze_field,
            },
            "highlight": {
                "style": self._default_highlight_style,
            },
        }

        self._config.read_dict(config_dict)
        self._save()

    def _read(self):
        """Get config from file system"""
        self._config.read(self._config_path)

    def _save(self):
        """Save config state in file system"""
        with open(self._config_path, mode="wt", encoding="utf-8") as file:
            self._config.write(file)

    def get_option_value(self, section: str, key: str) -> str:
        """Get value of the config option"""
        return self._config[section][key]

    def update_option_value(self, section: str, key: str, new_value: str):
        """Update value of the config option"""
        if key not in self._config[section]:
            raise KeyError

        self._config[section][key] = new_value
        self._save()

    def get_formatted_options(self) -> List[str]:
        """Get list of formatted key-value entries from the config"""
        formatted_entries = []
        for section in self._config.sections():
            for key, value in self._config[section].items():
                formatted_entries.append(f"{section}.{key} = {value}")

        return formatted_entries

    def __rich__(self) -> Table:
        table = Table(
            Column("Option", justify="left", style="magenta"),
            Column("Value", justify="left", style="green"),
        )
        for section in self._config.sections():
            for key, value in self._config[section].items():
                table.add_row(f"{section}.{key}", value, end_section=True)

        return table

    def __repr__(self):
        return f"{type(self).__name__}(config_path={repr(self._config_path)})"



import os

import pytest

from inka2.models.hasher import Hasher


@pytest.fixture
def test_file(tmp_path) -> str:
    path = str(tmp_path / "file.md")
    with open(path, mode="wt", encoding="utf-8") as f:
        f.write("Hello, **WORLD**!")
    yield path
    os.remove(path)


@pytest.fixture
def file_with_hashes(tmp_path) -> str:
    path = str(tmp_path / "test_file_with_hashes.json")
    with open(path, mode="wt", encoding="utf-8") as f:
        f.write(
            "{\n"
            '"fake1.md": "724ae4e3bf11c04581daf1bb4de43161",\n'
            '"fake2.md": "123ae4a4bf1ac04581daf1bb4de43161"\n'
            "}"
        )

    yield path
    os.remove(path)


@pytest.fixture
def hasher(file_with_hashes: str) -> Hasher:
    return Hasher(file_with_hashes)


# calculate_hash
def test_calculate_hash_when_file_exists(test_file):
    expected = "724ae4e3bf11c04581daf1bb4de43161"

    assert Hasher.calculate_hash(test_file) == expected


def test_calculate_hash_when_file_does_not_exist_raises_error():
    with pytest.raises(FileNotFoundError):
        Hasher.calculate_hash("does_not_exist.json")


# has_changed
def test_has_changed_when_file_hash_exists_in_data_file_and_hash_same(hasher):
    assert not hasher.has_changed("fake1.md", "724ae4e3bf11c04581daf1bb4de43161")


def test_has_changed_when_file_hash_exists_in_data_file_and_hashes_different(hasher):
    assert hasher.has_changed("fake1.md", "111ae1e31111c04581daf1bb4de43161")


def test_has_changed_when_file_hash_does_not_exist_in_data_file(hasher):
    assert hasher.has_changed("oops.md", "111ae1e31111c04581daf1bb4de43161")


# update_hash
def test_update_hash_when_file_hash_exists_in_data_file(hasher, file_with_hashes):
    new_hash = "111ae1e31111c04581daf1bb4de43161"
    expected = (
        "{"
        f'"fake1.md": "{new_hash}", '
        '"fake2.md": "123ae4a4bf1ac04581daf1bb4de43161"'
        "}"
    )

    hasher.update_hash("fake1.md", new_hash)

    with open(file_with_hashes, mode="rt", encoding="utf-8") as f:
        assert f.read() == expected


def test_update_hash_when_file_hash_does_not_exist_in_data_file(
    hasher, file_with_hashes
):
    filename = "my_file_path.md"
    new_hash = "111ae1e31111c04581daf1bb4de43161"
    expected = (
        "{"
        '"fake1.md": "724ae4e3bf11c04581daf1bb4de43161", '
        '"fake2.md": "123ae4a4bf1ac04581daf1bb4de43161", '
        f'"{filename}": "{new_hash}"'
        "}"
    )

    hasher.update_hash(filename, new_hash)

    with open(file_with_hashes, mode="rt", encoding="utf-8") as f:
        assert f.read() == expected


# reset_hashes
def test_reset_hashes_when_data_file_does_not_exist():
    path = "some_random_path.json"
    hasher = Hasher(path)
    expected = "{}"

    hasher.reset_hashes()

    with open(path, mode="rt", encoding="utf-8") as f:
        assert f.read() == expected
    os.remove(path)


def test_reset_hashes_when_data_file_contains_hashes(hasher, file_with_hashes):
    expected = "{}"

    hasher.reset_hashes()

    with open(file_with_hashes, mode="rt", encoding="utf-8") as f:
        assert f.read() == expected



# Parts of code were impudently borrowed from AnkiConnect(https://github.com/FooSoft/anki-connect)
# and apy (https://github.com/lervag/apy) projects. Thanks to all their developers.

import os
from pathlib import Path
from typing import Dict, Iterable, List, Type

import anki
import anki.collection
import anki.consts
import anki.errors
import anki.models
import anki.notes
from aqt.profiles import ProfileManager

from ..exceptions import AnkiApiError
from .config import Config
from .notes.note import Note


class AnkiApi:
    """Class for working with Anki collection"""

    def __init__(self, cfg: Config, anki_path: Path):
        self._cfg = cfg

        # Check correctness of the anki path
        meta_path = os.path.join(anki_path, "prefs21.db")
        if not os.path.exists(meta_path):
            raise AnkiApiError(f'incorrect path to Anki folder: "{anki_path}"')

        # Initialize profile manager to get access to profile and database actions
        self._profile_manager = ProfileManager(anki_path)
        self._profile_manager.setupMeta()

    def get_profiles(self) -> List[str]:
        """Get list of user profiles from Anki"""
        return self._profile_manager.profiles()

    def load_collection(self, profile: str) -> None:
        """Select profile in Anki and load collection"""
        # opening collection changes current directory so
        # we need to rollback it to initial
        initial_dir = os.getcwd()

        try:
            self._profile_manager.load(profile)
            self._collection = anki.collection.Collection(
                self._profile_manager.collectionPath()
            )
        except anki.errors.DBError:
            raise AnkiApiError(
                "the collection is open in Anki. "
                "You need to either close Anki or switch to a different profile."
            )

        os.chdir(initial_dir)

    def sync(self):
        """Sync collection to AnkiWeb"""
        # todo: handle case when to sync we need user decision (download or upload)
        auth = self._profile_manager.sync_auth()
        if auth is None:
            raise AnkiApiError("Isn't authenticated with AnkiWeb")

        # self._collection.save(trx=False)

        # Perform main sync
        try:
            # TODO it throws a warning that it is being running in the main thread
            # https://forums.ankiweb.net/t/unable-to-sync-login-via-anki-python-pkg-to-ankiweb/43092/4
            sync_media = self._profile_manager.media_syncing_enabled()
            self._collection.sync_collection(auth, sync_media)
        except anki.errors.NetworkError:
            raise AnkiApiError("Please check your internet connection")

        # Perform media sync
        initial_dir = os.getcwd()
        os.chdir(self._collection.media.dir())
        self._collection.sync_media(auth)
        os.chdir(initial_dir)

    def add_note(self, note: Note) -> int:
        model = self._collection.models.by_name(note.get_anki_note_type(self._cfg))
        anki_note = anki.notes.Note(self._collection, model)
        anki_note.tags = list(note.tags)
        anki_note.fields = list(note.get_html_fields(self._cfg).values())

        if anki_note.dupeOrEmpty():
            raise AnkiApiError("duplicate! Note wasn't added.", note=note)

        # gets id of the deck. if deck doesn't exist - it will be created.
        deck_id = self._collection.decks.id(note.deck_name, create=True)
        if not deck_id:
            raise AnkiApiError(f"the deck {note.deck_name} couldn't be created.")

        self._collection.add_note(anki_note, deck_id)
        return anki_note.id

    def update_note_ids(self, notes: Iterable[Note]) -> None:
        """Update incorrect or absent IDs of notes"""
        for note in notes:
            # Update id only if note with this id doesn't exist
            try:
                note_id = anki.notes.NoteId(note.anki_id if note.anki_id else -1)
                self._collection.get_note(note_id)
            except anki.errors.NotFoundError:
                found_notes = self._collection.find_notes(note.search_query)
                note.anki_id = found_notes[0] if found_notes else None

    def update_note(self, note: Note) -> None:
        """Synchronize changes in notes with Anki"""
        try:
            note_id = anki.notes.NoteId(note.anki_id if note.anki_id else -1)
            anki_note = self._collection.get_note(note_id)
        except anki.errors.NotFoundError:
            raise AnkiApiError(
                f"note with ID {note.anki_id} was not found. "
                f'You can update IDs on notes with the command "inka collect --update-ids path/to/file.md".'
            )

        for field, value in note.get_html_fields(self._cfg).items():
            if field in anki_note:
                anki_note[field] = value
        # anki_note.flush()  # TODO: deprecated warning:col.update_notes()
        # Use the collection's update_note method instead of the note's flush method
        self._collection.update_note(anki_note)

    def fetch_note_types(self) -> List[str]:
        """Get list of names of the existing note types"""
        return [n.name for n in self._collection.models.all_names_and_ids()]

    def create_note_type(
        self,
        name: str,
        fields: List[str],
        css: str,
        card_templates: List[Dict[str, str]],
        is_cloze: bool,
    ) -> None:
        """Create new note type"""
        model = self._collection.models.new(name)
        if is_cloze:
            model["type"] = anki.consts.MODEL_CLOZE

        # Create fields and add them to Note
        for field in fields:
            model_field = self._collection.models.new_field(field)
            self._collection.models.add_field(model, model_field)

        # Add shared css to model if exists. Use default otherwise
        if css:
            model["css"] = css

        # Generate new card template(s)
        counter = 1
        for card in card_templates:
            card_name = "Card " + str(counter)
            if "Name" in card:
                card_name = card["Name"]

            tmpl = self._collection.models.new_template(card_name)
            tmpl["qfmt"] = card["Front"]
            tmpl["afmt"] = card["Back"]
            self._collection.models.add_template(model, tmpl)
            counter += 1

        self._collection.models.add(model)

    def fetch_note_type_styling(self, note_type: Type[Note]) -> str:
        """Get styling of note type that is used to add notes"""
        return self._get_model(note_type)["css"]

    def update_note_type_styling(self, note_type: Type[Note], new_styles: str) -> None:
        """Update styling of the note type in Anki"""
        anki_model = self._get_model(note_type)
        anki_model["css"] = new_styles
        self._collection.models.update_dict(anki_model)

    def fetch_note_type_templates(
        self, note_type: Type[Note]
    ) -> Dict[str, Dict[str, str]]:
        """Get templates of note type"""
        model = self._get_model(note_type)

        templates = {}
        for template in model["tmpls"]:
            templates[template["name"]] = {
                "Front": template["qfmt"],
                "Back": template["afmt"],
            }

        return templates

    def update_note_type_templates(
        self, note_type: Type[Note], templates: Dict[str, Dict[str, str]]
    ) -> None:
        """Update note type templates"""
        anki_model = self._get_model(note_type)

        for anki_template in anki_model["tmpls"]:
            template = templates.get(anki_template["name"])
            if template:
                qfmt = template.get("Front")
                if qfmt:
                    anki_template["qfmt"] = qfmt

                afmt = template.get("Back")
                if afmt:
                    anki_template["afmt"] = afmt

        self._collection.models.update_dict(anki_model)

    def close(self):
        """Save and close the collection."""
        self._collection.close()

    def _get_model(self, note_type: Type[Note]) -> anki.models.NoteType:
        model_name = note_type.get_anki_note_type(self._cfg)
        anki_model = self._collection.models.by_name(model_name)
        if not anki_model:
            raise AnkiApiError(f"couldn't get note type {model_name} from Anki!")

        return anki_model



import filecmp
import os
import shutil

import pytest


def test_file_exists_returns_true(anki_media, image_anki):
    expected = True
    file_name = os.path.basename(image_anki)

    assert anki_media.exists(file_name) == expected


def test_file_does_not_exist_returns_false(anki_media, image):
    expected = False
    file_name = os.path.basename(image)

    assert anki_media.exists(file_name) == expected


def test_copying_non_existing_file_raises_error(anki_media):
    with pytest.raises(FileNotFoundError):
        anki_media.copy_file_from("path/does/not/exist.md")


def test_copying_file_when_no_file_with_same_name_in_anki_media(
    anki_media, image, path_to_anki_image
):
    expected = True

    anki_media.copy_file_from(image)

    assert filecmp.cmp(image, path_to_anki_image) == expected


def test_copying_file_when_exists_different_file_with_same_name_raises_error(
    anki_media, image, path_to_anki_image
):
    with open(path_to_anki_image, "wt") as f:
        f.write("some text")

    with pytest.raises(FileExistsError):
        anki_media.copy_file_from(image)


def test_copying_file_when_exists_same_file_with_same_name(anki_media, image):
    anki_file_path = f"{anki_media._anki_media_path}/{image}"
    shutil.copyfile(image, anki_file_path)

    anki_media.copy_file_from(image)

    assert filecmp.cmp(image, anki_file_path)


def test_creating_file_when_no_file_with_same_name_in_anki_media(anki_media):
    file_name = "_default.css"
    file_content = "body { background-color: yellow; }"
    anki_file_path = f"{anki_media._anki_media_path}/{file_name}"

    anki_media.create_file(file_name, file_content)

    with open(anki_file_path, "rt") as f:
        assert f.read() == file_content

    os.remove(anki_file_path)


def test_creating_file_when_exists_file_with_same_name_in_anki_media(anki_media):
    file_name = "_default.css"
    file_content = "body { background-color: yellow; }"
    anki_file_path = f"{anki_media._anki_media_path}/{file_name}"
    with open(anki_file_path, "wt") as f:
        f.write(file_content)

    with pytest.raises(FileExistsError):
        anki_media.create_file(file_name, file_content)

    os.remove(anki_file_path)



import logging
import os
import random
import sys
from unittest.mock import MagicMock

import mistune
import pytest
from PIL import Image as Img

from inka2.cli import ROOT_DIR
from inka2.helpers import parse_str_to_bool
from inka2.mistune_plugins.mathjax3 import plugin_mathjax3
from inka2.models.anki_api import AnkiApi
from inka2.models.anki_media import AnkiMedia
from inka2.models.config import Config
from inka2.models.parser import Parser

log_fmt = r"%(asctime)-15s %(levelname)s %(name)s %(funcName)s:%(lineno)d %(message)s"
datefmt = "%Y-%m-%d %H:%M:%S"
logging.basicConfig(format=log_fmt, level=logging.DEBUG, datefmt=datefmt)

DEFAULT_ANKI_FOLDERS = {
    "win32": r"~\AppData\Roaming\Anki2",
    "linux": "~/.local/share/Anki2",
    "darwin": "~/Library/Application Support/Anki2",
}


@pytest.fixture
def anki_media() -> AnkiMedia:
    """Instance of AnkiMedia class with profile 'test'."""
    anki_media = AnkiMedia(
        "test_profile", os.path.expanduser(DEFAULT_ANKI_FOLDERS[sys.platform])
    )

    try:
        os.makedirs(anki_media._anki_media_path)
    except FileExistsError:
        return anki_media

    return anki_media


@pytest.fixture
def anki_media_mock(mocker) -> MagicMock:
    """Mock of AnkiMedia instance"""
    return mocker.MagicMock(spec=AnkiMedia)


@pytest.fixture
def fake_parser(config: Config) -> Parser:
    """Parser class with dummy filename, default_deck. It uses 'test' profile."""
    return Parser(config, "file_doesnt_exist.md", "")


@pytest.fixture
def anki_api_mock(mocker) -> MagicMock:
    """Mock of AnkiApi instance"""
    return mocker.MagicMock(spec=AnkiApi)


@pytest.fixture
def image() -> str:
    """Path to temp image in working directory"""
    path = f"image_for_testing_{random.randrange(999999)}.png"
    Img.new("RGBA", size=(50, 50), color=(155, 0, 0)).save(path, format="png")

    yield path

    os.remove(path)


@pytest.fixture
def another_image() -> str:
    """Path to temp image in working directory"""
    path = f"another_image_for_testing_{random.randrange(999999)}.png"
    Img.new("RGBA", size=(50, 50), color=(155, 0, 123)).save(path, format="png")

    yield path

    os.remove(path)


@pytest.fixture
def path_to_anki_image(anki_media, image):
    """Path to non-existing image (with the same name as 'image' fixture returns) in anki media folder"""
    path_to_anki_image = f"{anki_media._anki_media_path}/{image}"

    yield path_to_anki_image

    # Remove image if it was created
    try:
        os.remove(path_to_anki_image)
    except OSError:
        pass


@pytest.fixture
def path_to_another_anki_image(anki_media, another_image):
    """Path to non-existing image (with the same name as 'another_image' fixture returns) in anki media folder"""
    path = f"{anki_media._anki_media_path}/{another_image}"

    yield path

    # Remove image if it was created
    try:
        os.remove(path)
    except OSError:
        pass


@pytest.fixture
def image_anki(anki_media, path_to_anki_image) -> str:
    """Temporary image in anki media folder with the same name as 'image' fixture returns
    but different content.
    """
    Img.new("RGBA", size=(20, 150), color=(100, 13, 35)).save(
        path_to_anki_image, format="png"
    )
    return path_to_anki_image


@pytest.fixture
def config_path(tmp_path):
    """Temporary path to config file"""
    return tmp_path / "test_config.ini"


@pytest.fixture
def config(config_path):
    """Instance of Config class. Path to config specified by 'config_path' fixture"""
    return Config(config_path)


@pytest.fixture
def config_add_filename() -> Config:
    return Config(f"{ROOT_DIR}/tests/resources/defaults/config_add_filename.ini")


@pytest.fixture
def md(config_add_filename):
    return mistune.create_markdown(
        plugins=["strikethrough", "footnotes", "table", plugin_mathjax3],
        escape=parse_str_to_bool(
            config_add_filename.get_option_value("defaults", "escape_html")
        ),
    )



import configparser
import os

import pytest

from inka2.models.config import Config


@pytest.fixture
def default_config_string():
    """Contents of the default config file"""
    return (
        "[defaults]\n"
        "profile = \n"
        "deck = Default\n"
        "folder = \n"
        "escape_html = False\n"
        "add_filename = False\n"
        "\n"
        "[anki]\n"
        "path = \n"
        "basic_type = Inka Basic\n"
        "front_field = Front\n"
        "back_field = Back\n"
        "cloze_type = Inka Cloze\n"
        "cloze_field = Text\n"
        "\n"
        "[highlight]\n"
        "style = monokai\n"
        "\n"
    )


def test_saves_correctly(tmp_path):
    filepath = tmp_path / "test_config.ini"
    config = Config(filepath)
    config._config = configparser.ConfigParser()
    config._config["defaults"] = {"deck": "Default", "folder": ""}
    expected = "[defaults]\n" "deck = Default\n" "folder = \n" "\n"

    config._save()

    with open(filepath, mode="rt", encoding="utf-8") as file:
        assert file.read() == expected


def test_reads_correctly_from_existing_config(config_path):
    with open(config_path, mode="wt", encoding="utf-8") as file:
        file.write(
            "[defaults]\n"
            "deck = Default\n"
            "folder =\n"
            "\n"
            "[anki_connect]\n"
            "port = 8765\n"
        )
    config = Config(config_path)

    expected = configparser.ConfigParser()
    expected["defaults"] = {"deck": "Default", "folder": ""}
    expected["anki_connect"] = {"port": "8765"}

    assert config._config == expected


def test_create_default_config(config, config_path, default_config_string):
    os.remove(config_path)

    config._create_default()

    with open(config_path, mode="rt", encoding="utf-8") as file:
        assert file.read() == default_config_string


def test_if_config_not_found_created_default(
    config, config_path, default_config_string
):
    with open(config_path, mode="rt", encoding="utf-8") as file:
        assert file.read() == default_config_string


def test_get_entry_value(config):
    expected = "Default"

    entry_value = config.get_option_value("defaults", "deck")

    assert entry_value == expected


def test_get_entry_value_with_incorrect_section_raises_error(config):
    with pytest.raises(KeyError):
        config.get_option_value("ankiConnect", "port")


def test_get_entry_value_with_incorrect_key_raises_error(config):
    with pytest.raises(KeyError):
        config.get_option_value("anki_connect", "profile")


def test_updates_entry_value_in_backing_object(config):
    new_value = "path/to/folder"
    config.update_option_value("defaults", "folder", new_value)

    updated_value = config._config["defaults"]["folder"]

    assert updated_value == new_value


def test_updates_entry_value_in_file(config, config_path):
    new_value = "path/to/folder"
    config.update_option_value("defaults", "folder", new_value)
    config_parser = configparser.ConfigParser()

    config_parser.read(config_path)
    updated_value = config_parser["defaults"]["folder"]

    assert updated_value == new_value


def test_update_option_value_with_incorrect_section_raises_error(config):
    with pytest.raises(KeyError):
        config.update_option_value("ankiConnect", "port", "12")


def test_update_option_value_with_incorrect_key_raises_error(config):
    with pytest.raises(KeyError):
        config.update_option_value("defaults", "port", "Profile Name")


def test_repr_method(config, config_path):
    expected = f"Config(config_path={repr(config_path)})"

    assert repr(config) == expected


def test_reset(config, config_path, default_config_string):
    config.update_option_value("defaults", "profile", "new profile")

    config.reset()

    with open(config_path, mode="rt", encoding="utf-8") as f:
        assert f.read() == default_config_string


def test_new_options_removed_by_reset(config, config_path, default_config_string):
    config._config = configparser.ConfigParser()
    new_config = {
        "defaults": {
            "profile": config._default_profile,
            "deck": config._default_deck,
            "folder": config._default_folder,
            "new_option": 123456,
        },
        "anki": {
            "note_type": config._default_basic_type,
            "front_field": config._default_front_field,
            "back_field": config._default_back_field,
        },
    }
    config._config.read_dict(new_config)
    config._save()

    config.reset()

    with open(config_path, mode="rt", encoding="utf-8") as f:
        assert f.read() == default_config_string



import pytest

test_cases = {
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): "Some question?",
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "12. Some question?\n"
        "\n"
        "> Answer"
    ): "Some question?",
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "123. Some question?\n"
        "\n"
        "> Answer"
    ): "Some question?",
    (
        "Deck: Abraham\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer"
    ): "Some question?\n\nMore info on question.\n\nAnd even more!",
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:123456-->\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): "Some question?",
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. A little bit of python code...\n"
        "\n"
        "```python\n"
        "def hello(name: str) -> str:\n"
        "    return f'Hello, {name}!'\n"
        "\n"
        "if __name__ == '__main__':\n"
        "    print(hello('bro'))\n"
        "```\n"
        "\n"
        "> some text"
    ): (
        "A little bit of python code...\n"
        "\n"
        "```python\n"
        "def hello(name: str) -> str:\n"
        "    return f'Hello, {name}!'\n"
        "\n"
        "if __name__ == '__main__':\n"
        "    print(hello('bro'))\n"
        "```"
    ),
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some text before:\n"
        "\n"
        "```python\n"
        "def hello(name: str) -> str:\n"
        "    return f'Hello, {name}!'\n"
        "\n"
        "if __name__ == '__main__':\n"
        "    print(hello('bro'))\n"
        "```\n"
        "\n"
        "text in between\n"
        "\n"
        "```python\n"
        "def hello(name: str) -> str:\n"
        "    return f'Hello, {name}!\n"
        "```\n"
        "\n"
        "text after\n"
        "\n"
        "```commandline\n"
        "inka collect -u path/to/file.md\n"
        "```\n"
        "\n"
        "> Short answer\n"
        "> here"
    ): (
        "Some text before:\n"
        "\n"
        "```python\n"
        "def hello(name: str) -> str:\n"
        "    return f'Hello, {name}!'\n"
        "\n"
        "if __name__ == '__main__':\n"
        "    print(hello('bro'))\n"
        "```\n"
        "\n"
        "text in between\n"
        "\n"
        "```python\n"
        "def hello(name: str) -> str:\n"
        "    return f'Hello, {name}!\n"
        "```\n"
        "\n"
        "text after\n"
        "\n"
        "```commandline\n"
        "inka collect -u path/to/file.md\n"
        "```"
    ),
    # cloze
    ("1. Some {question?}\n" "\n"): "Some {question?}",
    # cloze anki format
    ("2. Some {{c1::question?}}\n" "\n"): "Some {{c1::question?}}",
    # cloze with ID
    ("<!--ID:1612579125074-->\n" "32. Some {question?}\n" "\n"): "Some {question?}",
    # multiline cloze
    (
        "1. Some {question?}\n"
        "\n"
        "More {{c1::info on question}}.\n"
        "\n"
        "{1::And::hint} even more!"
    ): (
        "Some {question?}\n"
        "\n"
        "More {{c1::info on question}}.\n"
        "\n"
        "{1::And::hint} even more!"
    ),
    # no question
    "Some text": None,
}


@pytest.mark.parametrize("text, expected", test_cases.items())
def test_get_question(fake_parser, text, expected):
    question = fake_parser.get_question(text)

    assert question == expected



import re
from pathlib import Path
from typing import Iterable, List, Optional, Union

from ..helpers import print_sub_warning
from .notes.basic_note import BasicNote
from .notes.cloze_note import ClozeNote
from .notes.note import Note
from .parser import Parser


class Writer:
    """Class for editing file with notes"""

    def __init__(self, file_path: Union[str, Path], notes: Iterable[Note]):
        self._file_path = file_path
        self._notes = notes

        with open(self._file_path, mode="rt", encoding="utf-8") as f:
            self._file_content = f.read()
        self._note_strings = Parser.get_note_strings(self._file_content)

    def update_note_ids(self):
        """Update lines with IDs of the notes from the file"""
        for note in self._notes:
            # Find note's question in file string
            note_question = note.get_raw_question_field()

            all_occurrences = self.find_all_occurrences(note_question)
            if len(all_occurrences) > 1:
                print_sub_warning(
                    f"Warning: more than one occurrence of '{note_question}' found in {self._file_path}"
                )
                print_sub_warning("Can cause undefined behavior. Please fix.")

            question_string_start = self._file_content.find(
                note_question
            )  # TODO: make unambiguous
            if question_string_start == -1:
                continue

            matches = re.finditer(
                r"(<!--ID:\S+-->)?(\n\d+\.\s*)",
                self._file_content[:question_string_start],
                re.MULTILINE,
            )
            if not matches:
                # Note may not be found cause it was deleted
                continue

            # Search for existing line with ID
            question_match = list(matches)[-1]
            existing_id = Parser.get_id(question_match.group(0))

            # Skip if ID hasn't changed
            if existing_id == note.anki_id:
                continue

            id_string = f"<!--ID:{note.anki_id}-->\n"
            if not note.anki_id:
                # Delete ID from note in file if note object has no ID
                id_string = ""

            # index of newline character that comes before question
            newline_index = question_match.start(2)

            if existing_id is None:
                # Add line with ID after newline
                self._file_content = (
                    self._file_content[: newline_index + 1]
                    + id_string
                    + self._file_content[newline_index + 1 :]
                )
            else:
                # Substitute existing ID with the new one
                self._file_content = self._file_content.replace(
                    f"<!--ID:{existing_id}-->\n", id_string
                )

        self._save()

    def find_all_occurrences(self, note_question: str) -> List[int]:
        start = 0
        all_occurrences = []
        while True:
            start = self._file_content.find(note_question, start)
            if start == -1:  # No more occurrences
                break
            all_occurrences.append(start)
            start += len(note_question)  # Move past the current match
        return all_occurrences

    def update_fields_of_basic_notes(self):
        """Update question and answer fields in notes in file"""
        for note in self._notes:
            if not isinstance(note, BasicNote):
                continue

            if not note.changed:
                continue

            # Find string with this note by its ID
            note_string = self._get_note_string_by_id(note.anki_id)

            # Substitute question field
            current_question = Parser.get_question(note_string)
            self._file_content = self._file_content.replace(
                current_question, note.raw_front_md
            )

            # Create new answer field (with '>')
            current_answer = Parser.get_answer(note_string).rstrip()
            lines = note.raw_back_md.replace("\n\n", "\n").splitlines()
            new_answer = "\n".join(map(lambda line: f"> {line}", lines))

            # Substitute answer field
            self._file_content = self._file_content.replace(current_answer, new_answer)

        self._save()

    def delete_notes(self):
        """Delete notes marked for deletion from the file"""
        for note in self._notes:
            if not note.to_delete:
                continue

            # Find string with this note by its ID
            note_string = self._get_note_string_by_id(note.anki_id)

            # Delete note text from file
            self._file_content = self._file_content.replace(note_string, "")

        self._save()

    def update_cloze_notes(self):
        """Updates all cloze notes with the values from updated_text_md field"""
        for note in self._notes:
            if not isinstance(note, ClozeNote):
                continue

            self._file_content = self._file_content.replace(
                note.raw_text_md, note.updated_text_md, 1
            )
            note.raw_text_md = note.updated_text_md  # update info about raw text

        self._save()

    def _save(self):
        """Save file state into the file system"""
        with open(self._file_path, mode="wt", encoding="utf-8") as f:
            f.write(self._file_content)
        self._note_strings = Parser.get_note_strings(self._file_content)

    def _get_note_string_by_id(self, note_id: int) -> Optional[str]:
        """Gets note string from the file by its ID. If note wasn't found returns None"""
        for string in self._note_strings:
            if string.find(str(note_id)) != -1:
                return string
        return None

    def __repr__(self):
        return f"{type(self).__name__}(file_path={self._file_path!r}, notes={self._notes!r})"



import os

import pytest

from inka2.models import img_handler
from inka2.models.notes.basic_note import BasicNote
from inka2.models.notes.cloze_note import ClozeNote


def test_fetch_image_links_if_cards_do_not_have_images():
    card1 = BasicNote(
        front_md="Some text", back_md="Another text", tags=[], deck_name="deck"
    )
    card2 = BasicNote(
        front_md="More\n\nText", back_md="Answer text", tags=[], deck_name="deck"
    )
    card3 = ClozeNote(text_md="Some {{c1::question}} here", tags=[], deck_name="deck")

    links = img_handler._fetch_image_links([card1, card2, card3])

    assert {} == links


def test_fetch_image_links_if_cards_have_images():
    image1 = "![](path/to/img1.png)"
    image2 = "![](img2.png)"
    card1 = BasicNote(
        front_md="Some text",
        back_md=f"Another text\n\n{image1}",
        tags=[],
        deck_name="deck",
    )
    card2 = ClozeNote(text_md=f"{{More}}\n\n{image2}", tags=[], deck_name="deck")
    expected = {image1: [card1], image2: [card2]}

    links = img_handler._fetch_image_links([card1, card2])

    assert expected == links


def test_fetch_image_links_if_cards_have_same_image():
    image = "![](path/to/img1.png)"
    card1 = ClozeNote(
        text_md=f"long {{sentence}} or not\n\n{image}", tags=[], deck_name="deck"
    )
    card2 = BasicNote(
        front_md=f"More\n\n{image}", back_md="Answer text", tags=[], deck_name="deck"
    )
    expected = {image: [card1, card2]}

    links = img_handler._fetch_image_links([card1, card2])

    assert expected == links


def test_update_image_links_when_no_image_links():
    card1 = ClozeNote(text_md="Some text", tags=[], deck_name="deck")
    card2 = BasicNote(
        front_md="More\n\nText", back_md="Answer text", tags=[], deck_name="deck"
    )

    img_handler._update_image_links_in_notes({})

    assert card1.raw_text_md == card1.updated_text_md
    assert card2.raw_front_md == card2.updated_front_md
    assert card2.raw_back_md == card2.updated_back_md


def test_update_image_links_when_basic_card_has_image_link():
    image = "![](path/to/img1.png)"
    card = BasicNote(
        front_md=f"Some {image} text",
        back_md=f"Another text\n\n{image}",
        tags=[],
        deck_name="deck",
    )
    image_links = {image: [card]}
    expected_front = "Some ![](img1.png) text"
    expected_back = "Another text\n\n![](img1.png)"

    img_handler._update_image_links_in_notes(image_links)

    assert expected_front == card.updated_front_md
    assert expected_back == card.updated_back_md


def test_update_image_links_when_cloze_card_has_image_link():
    image = "![](path/to/img1.png)"
    card = ClozeNote(text_md=f"{{Some}} {image} text", tags=[], deck_name="deck")
    image_links = {image: [card]}
    expected = "{Some} ![](img1.png) text"

    img_handler._update_image_links_in_notes(image_links)

    assert card.updated_text_md == expected


def test_update_image_links_when_card_has_different_image_links():
    image1 = "![](path/to/img1.png)"
    image2 = "![](images/image2.png)"
    card = BasicNote(
        front_md=f"{image2}\n\nSome {image1} text",
        back_md=f"Another text\n\n{image1}\n{image2}",
        tags=[],
        deck_name="deck",
    )
    image_links = {image1: [card], image2: [card]}
    expected_front = "![](image2.png)\n\nSome ![](img1.png) text"
    expected_back = "Another text\n\n![](img1.png)\n![](image2.png)"

    img_handler._update_image_links_in_notes(image_links)

    assert expected_front == card.updated_front_md
    assert expected_back == card.updated_back_md


def test_update_image_links_in_several_basic_cards_with_same_images():
    image1 = "![](path/to/img1.png)"
    image2 = "![](images/image2.png)"
    card1 = BasicNote(
        front_md=f"{image2}\n\nSome {image1} text",
        back_md=f"Another text\n\n{image1}\n{image2}",
        tags=[],
        deck_name="deck",
    )
    card2 = BasicNote(
        front_md=f"{image2}Some text {image1}",
        back_md=f"{image1} {image2}\n\nAnother text\n\n",
        tags=[],
        deck_name="deck",
    )
    image_links = {image1: [card1, card2], image2: [card2, card1]}
    expected_front1 = "![](image2.png)\n\nSome ![](img1.png) text"
    expected_back1 = "Another text\n\n![](img1.png)\n![](image2.png)"
    expected_front2 = "![](image2.png)Some text ![](img1.png)"
    expected_back2 = "![](img1.png) ![](image2.png)\n\nAnother text\n\n"

    img_handler._update_image_links_in_notes(image_links)

    assert expected_front1 == card1.updated_front_md
    assert expected_back1 == card1.updated_back_md
    assert expected_front2 == card2.updated_front_md
    assert expected_back2 == card2.updated_back_md


def test_update_image_links_when_basic_and_cloze_notes():
    image1 = "![](path/to/img1.png)"
    image2 = "![](images/image2.png)"
    card1 = BasicNote(
        front_md=f"{image2}\n\nSome {image1} text",
        back_md=f"Another text\n\n{image1}\n{image2}",
        tags=[],
        deck_name="deck",
    )
    card2 = ClozeNote(
        text_md=f"{image2}Some text {{{image1}}}", tags=[], deck_name="deck"
    )
    image_links = {image1: [card1, card2], image2: [card2, card1]}
    expected_front = "![](image2.png)\n\nSome ![](img1.png) text"
    expected_back = "Another text\n\n![](img1.png)\n![](image2.png)"
    expected_cloze = "![](image2.png)Some text {![](img1.png)}"

    img_handler._update_image_links_in_notes(image_links)

    assert expected_front == card1.updated_front_md
    assert expected_back == card1.updated_back_md
    assert expected_cloze == card2.updated_text_md


# Copy images
def test_copy_images_to_with_one_image_link(anki_media, image, path_to_anki_image):
    img_handler._copy_images_to(anki_media, [f"![some image]({image})"])

    assert os.path.exists(path_to_anki_image)


def test_copy_images_to_with_multiple_image_links(
    anki_media, image, path_to_anki_image, another_image, path_to_another_anki_image
):
    img_handler._copy_images_to(
        anki_media, [f"![some image]({image})", f"![some image]({another_image})"]
    )

    assert os.path.exists(path_to_anki_image)
    assert os.path.exists(path_to_another_anki_image)


def test_copy_images_to_with_non_existing_image_raises_error(anki_media):
    with pytest.raises(FileNotFoundError):
        img_handler._copy_images_to(anki_media, ["![](/path/to/non-existing.png)"])


def test_copy_images_to_when_exists_different_image_with_same_name_raises_error(
    anki_media, image, image_anki
):
    with pytest.raises(FileExistsError):
        img_handler._copy_images_to(anki_media, [f"![some image]({image})"])


def test_handle_images_if_no_images_then_fields_remain_same(anki_media):
    card = BasicNote("Some text", "Another text", tags=[], deck_name="deck")

    img_handler.handle_images_in([card], anki_media)

    assert card.updated_front_md == card.raw_front_md


def test_handle_images_when_cards_have_images(anki_media, image, another_image):
    image_link1 = f"![some image]({image})"
    image_link2 = f"![another img]({another_image})"
    card1 = BasicNote(
        f"Some text {image_link1}",
        f"{image_link2} Another text",
        tags=[],
        deck_name="deck",
    )
    card2 = BasicNote(
        "Question here", f"And answer here {image_link2}", tags=[], deck_name="deck"
    )
    expected_front1 = f"Some text ![some image]({os.path.basename(image)})"
    expected_front2 = "Question here"
    expected_back1 = f"![another img]({os.path.basename(another_image)}) Another text"
    expected_back2 = (
        f"And answer here ![another img]({os.path.basename(another_image)})"
    )

    img_handler.handle_images_in([card1, card2], anki_media)

    assert expected_front1 == card1.raw_front_md
    assert expected_front2 == card2.raw_front_md
    assert expected_back1 == card1.raw_back_md
    assert expected_back2 == card2.raw_back_md


def test_handle_images_copies_images_to_anki_media(
    anki_media, image, path_to_anki_image
):
    card = BasicNote(
        "Some text", f"Another text ![img]({image})", tags=[], deck_name="deck"
    )

    img_handler.handle_images_in([card], anki_media)

    os.path.exists(path_to_anki_image)



import pytest

test_cases = {
    ("1. Question?\n" "\n" "Answer\n"): "Default",
    ("Deck: yolo\n" "1. Question?\n" "\n" "Answer\n"): "yolo",
    (
        "Deck: Very Long Deck Name\n" "1. Question?\n" "\n" "Answer\n"
    ): "Very Long Deck Name",
    (
        "1. Question?\n" "\n" "Answer\n" "\n" "Deck: yolo\n" "\n" "2. Q?\n" "\n" "A\n"
    ): "yolo",
    ("Some text; Deck: yolo\n" "1. Question?\n" "\n" "Answer\n"): "Default",
}

test_cases_error = [
    ("Some text; Deck: yolo\n" "1. Question?\n" "\n" "Answer\n"),
    ("Deck: Abraham\n" "1. Question?\n" "\n" "Deck: Default\n" "Answer\n"),
    ("1. Question?\n" "\n" "Answer\n"),
    ("Deck:\n" "1. Question?\n" "\n" "Answer\n"),
    ("Deck:   \n" "1. Question?\n" "\n" "Answer\n"),
]


@pytest.mark.parametrize("section, expected", test_cases.items())
def test_get_deck_name(fake_parser, section, expected):
    fake_parser._default_deck = "Default"

    deck = fake_parser._get_deck_name(section)

    assert expected == deck


@pytest.mark.parametrize("section", test_cases_error)
def test_get_deck_name_raises_error(fake_parser, section):
    with pytest.raises(ValueError):
        fake_parser._get_deck_name(section)



import pytest

from inka2.models.notes.basic_note import BasicNote
from inka2.models.notes.cloze_note import ClozeNote


@pytest.fixture
def cloze_note() -> ClozeNote:
    return ClozeNote("text content", ["tag1", "tag2"], "deck name")


def test_search_query(cloze_note):
    cloze_note.text_html = "<p>front content</p>"
    expected = '"<p>front content</p>"'

    assert cloze_note.search_query == expected


def test_convert_fields_to_html_when_function_passed(cloze_note, md):
    new_text = "new text"

    cloze_note.convert_fields_to_html(lambda text, md: new_text, md)

    assert cloze_note.text_html == new_text


def test_update_fields_with_when_function_passed(cloze_note):
    new_text = "new text"

    cloze_note.update_fields_with(lambda text: new_text)

    assert cloze_note.updated_text_md == new_text


def test_get_raw_fields(cloze_note):
    fields = cloze_note.get_raw_fields()

    assert len(fields) == 1
    assert fields[0] == cloze_note.raw_text_md


def test_get_raw_question_field(cloze_note):
    field = cloze_note.get_raw_question_field()

    assert field == cloze_note.raw_text_md


def test_get_html_fields(cloze_note, config):
    text_field = "myFront"
    config.update_option_value("anki", "cloze_field", text_field)
    expected = {
        text_field: cloze_note.text_html,
    }

    assert cloze_note.get_html_fields(config) == expected


def test_get_anki_note_type(cloze_note, config):
    expected = "my super type"
    config.update_option_value("anki", "cloze_type", expected)

    assert cloze_note.get_anki_note_type(config) == expected


def test_eq_when_same():
    first_note = ClozeNote("front", ["tag1"], "deck")
    second_note = ClozeNote("front", ["tag1"], "deck")

    assert first_note == second_note


@pytest.mark.parametrize(
    "second_note",
    (
        ClozeNote("oops", ["tag1"], "deck"),
        ClozeNote("front", ["tag1", "tag2"], "deck"),
        ClozeNote("front", ["tag1"], "my deck"),
        None,
        BasicNote("front", "back", ["tag1"], "my deck"),
        "short string",
    ),
)
def test_eq_when_not_equal(second_note):
    first_note = ClozeNote("front", ["tag1"], "deck")

    assert first_note != second_note



import pytest

test_cases = {
    # basic
    ("1. Some question?\n" "\n" "> Answer"): False,
    # basic with ID
    ("<!--ID:1612509025074-->\n" "1. Some question?\n" "\n" "> Answer"): False,
    # basic multiline question
    (
        "1. Some question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer"
    ): False,
    # basic multiline answer
    (
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> \n"
        "> And more to it"
    ): False,
    # basic multiline question and answer
    (
        "1. Some question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> \n"
        "> And more to it"
    ): False,
    # cloze
    ("1. Some {question?}\n" "\n"): True,
    # cloze anki format
    ("2. Some {{c1::question?}}\n" "\n"): True,
    # cloze with ID
    ("<!--ID:1612579125074-->\n" "32. Some {question?}\n" "\n"): True,
    # multiline cloze
    (
        "1. Some {question?}\n"
        "\n"
        "More {{c1::info on question}}.\n"
        "\n"
        "{1::And::hint} even more!"
    ): True,
    # not basic and not cloze
    ("3. Some question?\n" "\n"): False,
    # not basic and not cloze with ID
    ("<!--ID:1112809025074-->\n" "3. Some question?\n" "\n"): False,
    # basic with curly braces
    ("1. Some {question}?\n" "\n" "> Answer"): True,
    # empty
    "": False,
}


@pytest.mark.parametrize("string, expected", test_cases.items())
def test_is_cloze_note_str(fake_parser, string, expected):
    result = fake_parser._is_cloze_note_str(string)

    assert result == expected



import mistune
import pytest

from inka2.cli import CONFIG
from inka2.helpers import parse_str_to_bool
from inka2.mistune_plugins.mathjax3 import plugin_mathjax3
from inka2.models import converter
from inka2.models.notes.basic_note import BasicNote
from inka2.models.notes.cloze_note import ClozeNote


@pytest.fixture
def md():
    return mistune.create_markdown(
        plugins=["strikethrough", "footnotes", "table", plugin_mathjax3],
        escape=parse_str_to_bool(CONFIG.get_option_value("defaults", "escape_html")),
    )


@pytest.fixture
def basic_note():
    """BasicNote object with dummy data"""
    return BasicNote(front_md="dummy", back_md="dummy", tags=[], deck_name="")


convert_cloze_test_cases = {
    # SHORT IMPLICIT
    # one
    "Some {cloze question}": "Some {{c1::cloze question}}",
    # with hint
    "Some {cloze question::really helpful hint}": "Some {{c1::cloze question::really helpful hint}}",
    # two
    "Some {cloze question}\n\n{another} one": "Some {{c1::cloze question}}\n\n{{c2::another}} one",
    # around inline math
    r"math: {$\sqrt{2}$}": r"math: {{c1::$\sqrt{2}$}}",
    # around block math
    r"math: {$$\frac{\sqrt{2}}{15}$$}": r"math: {{c1::$$\frac{\sqrt{2}}{15}$$}}",
    # around inline code
    r"inline code: {`func() int { return {12} }`}": r"inline code: {{c1::`func() int { return {12} }`}}",
    # around block code
    (
        "Code:\n"
        "{\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
        "}\n"
    ): (
        "Code:\n"
        "{{c1::\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
        "}}\n"
    ),
    # check order correctness with multiple cloze deletions around math and code
    (
        r"some math {$\sqrt{2}$}, code {`func() int`}, "
        r"more code {`defer resp.body.Close()`} and more math {$\frac{1}{5}$}"
    ): (
        r"some math {{c1::$\sqrt{2}$}}, code {{c2::`func() int`}},"
        r" more code {{c3::`defer resp.body.Close()`}} and more math {{c4::$\frac{1}{5}$}}"
    ),
    # inside inline math -> ignored
    r"math: $\sqrt{2}$": r"math: $\sqrt{2}$",
    # inside block math -> ignored
    r"math: $$\frac{\sqrt{2}}{15}$$": r"math: $$\frac{\sqrt{2}}{15}$$",
    # inside inline code -> ignored
    r"inline code: `func() int { return {12} }`": r"inline code: `func() int { return {12} }`",
    # inside code block -> ignored
    (
        "Code:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ): (
        "Code:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ),
    # SHORT EXPLICIT
    # one
    "Some {1::cloze question}": "Some {{c1::cloze question}}",
    # with hint
    "Some {1::cloze question::really helpful hint}": "Some {{c1::cloze question::really helpful hint}}",
    # with 'c'
    "Some {c1::cloze question}": "Some {{c1::cloze question}}",
    # two
    "Some {2::cloze question}\n\n{1::another} one": "Some {{c2::cloze question}}\n\n{{c1::another}} one",
    # around inline math
    r"math: {1::$\sqrt{2}$}": r"math: {{c1::$\sqrt{2}$}}",
    # around block math
    r"math: {1::$$\frac{\sqrt{2}}{15}$$}": r"math: {{c1::$$\frac{\sqrt{2}}{15}$$}}",
    # around inline code
    r"inline code: {1::`func() int { return {12} }`}": r"inline code: {{c1::`func() int { return {12} }`}}",
    # around block code
    (
        "Code:\n"
        "{1::\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
        "}\n"
    ): (
        "Code:\n"
        "{{c1::\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
        "}}\n"
    ),
    # inside inline math -> ignored
    r"math: $\sqrt{2::15}$": r"math: $\sqrt{2::15}$",
    # inside block math -> ignored
    r"math: $$\frac{1::\sqrt{2}}{12::15}$$": r"math: $$\frac{1::\sqrt{2}}{12::15}$$",
    # inside inline code -> ignored
    r"inline code: `func() int { return {1::12} }`": r"inline code: `func() int { return {1::12} }`",
    # inside code block -> ignored
    (
        "Code:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("{1::exited}") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ): (
        "Code:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("{1::exited}") }\n'
        '\tdatabase.DB, err := sql.Open("pgx", os.Getenv("POSTGRES_URL"))\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ),
    # ANKI FORMAT
    # one
    "Some {{c1::cloze question}}": "Some {{c1::cloze question}}",
    # with hint
    "Some {{c1::cloze question::really helpful hint}}": "Some {{c1::cloze question::really helpful hint}}",
    # two
    "Some {{c2::cloze question}}\n\n{{c1::another}} one": "Some {{c2::cloze question}}\n\n{{c1::another}} one",
    # inside inline math -> ignored
    r"math: $\sqrt{ {{c1::15}} } $": r"math: $\sqrt{ {{c1::15}} } $",
    # inside block math -> ignored
    r"math: $$\frac{ {{1::\sqrt{2}}} }{ {{c12::15}} }$$": r"math: $$\frac{ {{1::\sqrt{2}}} }{ {{c12::15}} }$$",
    # block math and inline math -> ignored
    (
        r"math: $\sqrt{ {{c1::15}} } $ $$\frac{ {{1::\sqrt{2}}} }{ {{c12::15}} }$$"
    ): r"math: $\sqrt{ {{c1::15}} } $ $$\frac{ {{1::\sqrt{2}}} }{ {{c12::15}} }$$",
    # inside inline code -> ignored
    r"inline code: `func() int { return {{c1::12}} }`": r"inline code: `func() int { return {{c1::12}} }`",
    # inside code block -> ignored
    (
        "Code:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := {{c1::sql.Open("pgx", os.Getenv("POSTGRES_URL"))}}\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ): (
        "Code:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := {{c1::sql.Open("pgx", os.Getenv("POSTGRES_URL"))}}\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ),
    # inline and block code -> ignored
    (
        "Code of `{simple} {{c2::function}}`:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := {{c1::sql.Open("pgx", os.Getenv("POSTGRES_URL"))}}\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ): (
        "Code of `{simple} {{c2::function}}`:\n"
        "```\n"
        "func main() {\n"
        '\tdefer func() { fmt.Println("exited") }\n'
        '\tdatabase.DB, err := {{c1::sql.Open("pgx", os.Getenv("POSTGRES_URL"))}}\n'
        "\tif err != nil {\n"
        "\t\tlog.Fatal(err)\n"
        "\t}\n"
        "}\n"
        "```\n"
    ),
    # ALL THREE
    # random stuff
    (
        r"{Paris} is the capital and most populous city of {2::France},"
        r" with a estimated population of {2,148,271} residents"
    ): (
        r"{{c1::Paris}} is the capital and most populous city of {{c2::France}},"
        r" with a estimated population of {{c3::2,148,271}} residents"
    ),
    (
        r"{Paris} is the capital and most populous city of {3::France},"
        r" with a estimated population of {2,148,271} residents"
    ): (
        r"{{c1::Paris}} is the capital and most populous city of {{c3::France}},"
        r" with a estimated population of {{c3::2,148,271}} residents"
    ),
    (
        r"{c3::Paris} is the capital and most populous city of {3::France},"
        r" with a {{c1::estimated::my hint}} population of {2,148,271} residents"
    ): (
        r"{{c3::Paris}} is the capital and most populous city of {{c3::France}},"
        r" with a {{c1::estimated::my hint}} population of {{c4::2,148,271}} residents"
    ),
}

"""
refactoring of mathjax3.py results in the following edge case differences.
I am not aware of a clear specification what is correct.
mathjax3.py basically takes the builtin "math" plugin and changes only the delimiters.
"""
md_to_html_test_cases = {
    "some text here": "<p>some text here</p>",
    (
        "1. Item1\n"
        "2. Item2\n"
        "3. Item3\n"
        "\n"
        "```javascript\n"
        "let a = 12;\n"
        "let b = a;\n"
        "```\n"
        "\n"
        "[google](https://google.com)\n"
    ): (
        "<ol><li>Item1</li><li>Item2</li><li>Item3</li></ol>"
        '<pre><code class="language-javascript">let a = 12;\nlet b = a;</code></pre>'
        '<p><a href="https://google.com">google</a></p>'
    ),
    "some text here\n" "and more text": "<p>some text here\nand more text</p>",
    "some text here\n" "\n" "more text": "<p>some text here</p><p>more text</p>",
    # # inline mathjax
    r"$\sqrt{5}$": r'<p><span class="math">\(\sqrt{5}\)</span></p>',
    r"$\sqrt{5} $": r'<p><span class="math">\(\sqrt{5} \)</span></p>',
    # r"$ \sqrt{5}$": r"<p>\( \sqrt{5}\)</p>",  # TODO: fix this
    r"$multiple words$": r'<p><span class="math">\(multiple words\)</span></p>',
    r"weird$mathjax$more word$1$s": r'<p>weird<span class="math">\(mathjax\)</span>more word<span class="math">\(1\)</span>s</p>',
    "$\\sqrt{5}\n$": "<p>$\\sqrt{5}\n$</p>",
    "$\n\\sqrt{5}$": "<p>$\n\\sqrt{5}$</p>",
    r"\$\sqrt{5}$": r"<p>$\sqrt{5}$</p>",
    # r"$\sqrt{5}\$": r"<p>$\sqrt{5}$</p>",  # TODO: fix this
    r"\$\sqrt{5}\$": r"<p>$\sqrt{5}$</p>",
    r"text $$ here": r"<p>text $$ here</p>",
    # r"\$$$$": r"<p>$\(\)$</p>",  # TODO: fix this
    r"\$$\sqrt{2}$$": r'<p>$<span class="math">\(\sqrt{2}\)</span>$</p>',
    # r"$\$\sqrt{2}$$": r"<p>\(\$\sqrt{2}\)$</p>",  # TODO: fix this
    # r"$$\sqrt{2}\$$": r"<p>\(\)\sqrt{2}$$</p>",  # TODO: fix this
    # r"$$\sqrt{2}$\$": r"<p>\(\)\sqrt{2}$$</p>",  # TODO: fix this
    # r"\$\$\sqrt{2}$$": r"<p>$$\sqrt{2}\(\)</p>",  # TODO: fix this
    # r"$$\sqrt{2}\$\$": r"<p>\(\)\sqrt{2}$$</p>",  # TODO: fix this
    "$$$": r'<p><span class="math">\($\)</span></p>',
    # block mathjax
    # "$$$$": r"<p>\[\]</p>",  # TODO: wrong?
    # r"$$\sqrt{2}$$": r"<p>\[\sqrt{2}\]</p>",  # TODO: wrong?
    # r"inside $$\sqrt{2}$$ text": r"<p>inside \[\sqrt{2}\] text</p>",  # TODO: wrong?
    # r"$$ text here $$": r"<p>\[ text here \]</p>",  # TODO: wrong?
    # "$$multi\nline$$": "<p>\\[multi\nline\\]</p>",  # TODO: wrong?
    "$$\n\\sqrt{2}\n\\frac{1}{2}\n$$": '<div class="math">\\[\n\\sqrt{2}\n\\frac{1}{2}\n\\]</div>',
    # "$$\\sqrt{2}$$ some text in between $$\n\\sqrt{2}\n\\frac{1}{2}\n$$": (
    #     "<p>\\[\\sqrt{2}\\] some text in between \\[\n\\sqrt{2}\n\\frac{1}{2}\n\\]</p>"
    # ),  # TODO: unclear what should be correct
    # both inline and block mathjax
    # "$$\\sqrt{2}$$ some text $\\sqrt{6}$ in between $$\n\\sqrt{2}\n\\frac{1}{2}\n$$": (
    #     "<p>\\[\\sqrt{2}\\] some text \\(\\sqrt{6}\\) in between \\[\n\\sqrt{2}\n\\frac{1}{2}\n\\]</p>"
    # ),
}


@pytest.mark.parametrize("test_input, expected", convert_cloze_test_cases.items())
def test_convert_cloze_deletions_to_anki_format(test_input, expected):
    cloze_note = ClozeNote(text_md=test_input, tags=[], deck_name="")

    converter.convert_cloze_deletions_to_anki_format([cloze_note])

    assert cloze_note.updated_text_md == expected


def test_convert_cloze_deletions_to_anki_format_works_with_multiple_notes():
    text1 = (
        r"{Paris} is the capital and most populous city of {2::France},"
        r" with a estimated population of {2,148,271} residents"
    )
    text2 = "Some {{c1::cloze question}}"
    cloze_note1 = ClozeNote(text_md=text1, tags=[], deck_name="")
    cloze_note2 = ClozeNote(text_md=text2, tags=[], deck_name="")
    expected1 = (
        r"{{c1::Paris}} is the capital and most populous city of {{c2::France}},"
        r" with a estimated population of {{c3::2,148,271}} residents"
    )
    expected2 = "Some {{c1::cloze question}}"

    converter.convert_cloze_deletions_to_anki_format([cloze_note1, cloze_note2])

    assert cloze_note1.updated_text_md == expected1
    assert cloze_note2.updated_text_md == expected2


@pytest.mark.parametrize("test_input, expected", md_to_html_test_cases.items())
def test_convert_md_to_html(basic_note, test_input, expected, md):
    html = converter._convert_md_to_html(test_input, md)

    assert html == expected


def test_convert_md_to_html_no_html_escaping(basic_note, md):
    test_input = (
        """<span style="font-size: 9pt;">This text will be 9pt in size.</span>"""
    )
    expected = (
        """<p><span style="font-size: 9pt;">This text will be 9pt in size.</span></p>"""
    )
    html = converter._convert_md_to_html(test_input, md)
    assert html == expected


def test_convert_cards_to_html_works_with_multiple_cards(md):
    text1 = r"inside $\sqrt{2}$ text"
    text2 = "1. Item1\n" "2. Item2\n" "3. Item3\n"
    basic_note1 = BasicNote(front_md=text1, back_md=text2, tags=[], deck_name="")
    basic_note2 = BasicNote(front_md=text2, back_md=text1, tags=[], deck_name="")
    cloze_note = ClozeNote(text_md="Some question {{c1::42}}", tags=[], deck_name="")
    expected1 = r'<p>inside <span class="math">\(\sqrt{2}\)</span> text</p>'
    expected2 = "<ol><li>Item1</li><li>Item2</li><li>Item3</li></ol>"
    expected3 = "<p>Some question {{c1::42}}</p>"

    converter.convert_notes_to_html([basic_note1, basic_note2, cloze_note], md)

    assert basic_note1.front_html == expected1
    assert basic_note1.back_html == expected2
    assert basic_note2.front_html == expected2
    assert basic_note2.back_html == expected1
    assert cloze_note.text_html == expected3



import pytest

test_cases = {
    "": [],
    ("First line" "Second line" "\t 123 51"): [],
    ("---\n" "---"): [],
    ("---123\n" "Not empty\n" "---"): [],
    ("---\n" "Not empty\n" "a---"): [],
    ("---\n" "Not empty\n" "Ok"): [],
    ("---\n" "Text inside section\n" "---"): ["Text inside section\n"],
    ("---\n" "First\n" "Second\n" "Third\n" "---"): ["First\nSecond\nThird\n"],
    ("---\n" "First one\n" "---\n" "---\n" "Second one\n" "---"): [
        "First one\n",
        "Second one\n",
    ],
    (
        "---\n"
        "First one\n"
        "---\n"
        "Some text in between\n"
        "---\n"
        "Second one\n"
        "---"
    ): [
        "First one\n",
        "Second one\n",
    ],
}


@pytest.mark.parametrize("text, expected", test_cases.items())
def test_get_sections(fake_parser, text, expected):
    sections = fake_parser._get_sections(text)

    assert sections == expected



import pytest
import requests
from requests import HTTPError

from inka2.models import highlighter
from inka2.models.highlighter import BASE_URL
from inka2.models.notes.note import Note


@pytest.fixture
def default_styles() -> str:
    return """.card {
  font-family: arial;
  font-size: 20px;
  color: black;
  background-color: white;
}
"""


def test_update_style_in_note_type_when_style_is_empty_string_raises_error(
    anki_api_mock,
):
    with pytest.raises(ValueError):
        highlighter._update_style_in(
            note_type=Note, style_name="", anki_api=anki_api_mock
        )


def test_update_style_in_note_type_uses_anki_api_to_get_current_style(
    anki_api_mock, default_styles, mocker
):
    anki_api_mock.fetch_note_type_styling.return_value = default_styles
    mocker.patch("requests.get", return_value=mocker.MagicMock())

    highlighter._update_style_in(
        note_type=Note, style_name="monokai", anki_api=anki_api_mock
    )

    assert anki_api_mock.fetch_note_type_styling.call_count == 1


def test_update_style_in_note_type_uses_requests_to_download_style(
    anki_api_mock, mocker, default_styles
):
    anki_api_mock.fetch_note_type_styling.return_value = default_styles
    style_name = "monokai"
    mocked_get = mocker.patch("requests.get", return_value=mocker.MagicMock())
    url = f"{BASE_URL}/styles/{style_name}.min.css"

    highlighter._update_style_in(
        note_type=Note, style_name=style_name, anki_api=anki_api_mock
    )

    mocked_get.assert_called_once_with(url)


def test_update_style_in_note_type_skips_when_same_style_already_exists(
    anki_api_mock, mocker, default_styles
):
    mocker.patch("requests.get", return_value=mocker.MagicMock())
    style_name = "monokai"
    start = f"/*STYLE:{style_name} VERSION:{highlighter.HLJS_VERSION}*/"
    styles = f"{default_styles}\n{start}\nsomething here\n/*END*/"
    anki_api_mock.fetch_note_type_styling.return_value = styles

    highlighter._update_style_in(
        note_type=Note, style_name=style_name, anki_api=anki_api_mock
    )

    assert anki_api_mock.update_note_type_styling.call_count == 0


def test_update_style_in_note_type_when_incorrect_style_requested_raises_error(
    anki_api_mock, default_styles
):
    anki_api_mock.fetch_note_type_styling.return_value = default_styles

    with pytest.raises(HTTPError):
        highlighter._update_style_in(
            note_type=Note, style_name="i_dont_exist", anki_api=anki_api_mock
        )


def test_update_style_in_note_type_when_no_highlight_style_generates_new_with_one(
    anki_api_mock, mocker, default_styles
):
    note_type = Note
    anki_api_mock.fetch_note_type_styling.return_value = default_styles
    some_style = ".hljs{display:block;overflow-x:auto;padding:.5em;background:#272822;color:#ddd}"
    response_mock = mocker.MagicMock(spec=requests.Response)
    response_mock.text = some_style
    mocker.patch("requests.get", return_value=response_mock)
    start = f"/*STYLE:monokai VERSION:{highlighter.HLJS_VERSION}*/"
    expected = f"{default_styles}\n{start}\n{some_style}\n/*END*/"

    highlighter._update_style_in(
        note_type=note_type, style_name="monokai", anki_api=anki_api_mock
    )

    anki_api_mock.update_note_type_styling.assert_called_once_with(note_type, expected)


def test_update_style_in_note_type_when_exists_different_highlight_style_replaces_it(
    anki_api_mock, mocker, default_styles
):
    anki_api_mock.fetch_note_type_styling.return_value = (
        f"{default_styles}\n/*STYLE:default VERSION:{highlighter.HLJS_VERSION}*/\n"
        f"some default style here\n"
        f"/*END*/"
    )
    style_name = "monokai"
    note_type = Note
    some_style = ".hljs{display:block;overflow-x:auto;padding:.5em;background:#272822;color:#ddd}"
    response_mock = mocker.MagicMock(spec=requests.Response)
    response_mock.text = some_style
    mocker.patch("requests.get", return_value=response_mock)
    start = f"/*STYLE:monokai VERSION:{highlighter.HLJS_VERSION}*/"
    expected = f"{default_styles}\n{start}\n{some_style}\n/*END*/"

    highlighter._update_style_in(
        note_type=note_type, style_name=style_name, anki_api=anki_api_mock
    )

    anki_api_mock.update_note_type_styling.assert_called_once_with(note_type, expected)


def test_handle_highlighjs_script_when_no_script_in_anki_media_downloads_one(
    anki_api_mock, anki_media_mock, mocker
):
    anki_media_mock.exists.return_value = False
    response_mock = mocker.MagicMock(spec=requests.Response)
    file_content = "let a = 12;"
    response_mock.text = file_content
    mocked_get = mocker.patch("requests.get", return_value=response_mock)
    expected_url = f"{highlighter.BASE_URL}/highlight.min.js"

    highlighter._handle_highlighjs_files_for(Note, anki_media_mock, anki_api_mock)

    mocked_get.assert_called_once_with(expected_url)
    anki_media_mock.create_file.assert_called_once_with("_hl.pack.js", file_content)


def test_handle_highlighjs_script_when_script_exists_in_anki_media_does_not_download_one(
    anki_api_mock, anki_media_mock, mocker
):
    anki_media_mock.exists.return_value = True
    response_mock = mocker.MagicMock(spec=requests.Response)
    mocked_get = mocker.patch("requests.get", return_value=response_mock)

    highlighter._handle_highlighjs_files_for(Note, anki_media_mock, anki_api_mock)

    mocked_get.assert_not_called()
    anki_media_mock.assert_not_called()


def test_handle_highlighjs_script_adds_scripts_to_front_and_back_of_note_type(
    anki_api_mock, anki_media_mock
):
    note_type = Note
    anki_media_mock.exists.return_value = True
    anki_api_mock.fetch_note_type_templates.return_value = {
        "Card 1": {
            "Front": "{{Front}}",
            "Back": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}",
        }
    }
    script_elements = (
        f'<script src="_hl.pack.js"></script>\n{highlighter.AUTOSTART_SCRIPT}'
    )
    expected = {
        "Card 1": {
            "Front": "{{Front}}" + f"\n{script_elements}",
            "Back": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}"
            + f"\n{script_elements}",
        }
    }

    highlighter._handle_highlighjs_files_for(note_type, anki_media_mock, anki_api_mock)

    anki_api_mock.update_note_type_templates.assert_called_once_with(
        note_type, expected
    )


def test_handle_highlighjs_script_when_front_of_note_type_has_scripts_does_not_add_more_to_it(
    anki_api_mock, anki_media_mock
):
    note_type = Note
    anki_media_mock.exists.return_value = True
    script_elements = (
        f'<script src="_hl.pack.js"></script>\n{highlighter.AUTOSTART_SCRIPT}'
    )
    front_value = "{{Front}}" + f"\n{script_elements}"
    anki_api_mock.fetch_note_type_templates.return_value = {
        "Card 1": {
            "Front": front_value,
            "Back": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}",
        }
    }
    expected = {
        "Card 1": {
            "Front": front_value,
            "Back": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}"
            + f"\n{script_elements}",
        }
    }

    highlighter._handle_highlighjs_files_for(note_type, anki_media_mock, anki_api_mock)

    anki_api_mock.update_note_type_templates.assert_called_once_with(
        note_type, expected
    )


def test_handle_highlighjs_script_when_back_of_note_type_has_scripts_does_not_add_more_to_it(
    anki_api_mock, anki_media_mock
):
    note_type = Note
    anki_media_mock.exists.return_value = True
    script_elements = (
        f'<script src="_hl.pack.js"></script>\n{highlighter.AUTOSTART_SCRIPT}'
    )
    back_value = "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}" + f"\n{script_elements}"
    anki_api_mock.fetch_note_type_templates.return_value = {
        "Card 1": {"Front": "{{Front}}", "Back": back_value}
    }
    expected = {
        "Card 1": {"Front": "{{Front}}" + f"\n{script_elements}", "Back": back_value}
    }

    highlighter._handle_highlighjs_files_for(note_type, anki_media_mock, anki_api_mock)

    anki_api_mock.update_note_type_templates.assert_called_once_with(
        note_type, expected
    )


def test_handle_highlighjs_script_if_fields_have_not_change_no_update_request_sent(
    anki_api_mock, anki_media_mock
):
    anki_media_mock.exists.return_value = True
    script_elements = (
        f'<script src="_hl.pack.js"></script>\n{highlighter.AUTOSTART_SCRIPT}'
    )
    anki_api_mock.fetch_note_type_templates.return_value = {
        "Card 1": {
            "Front": "{{Front}}" + f"\n{script_elements}",
            "Back": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}"
            + f"\n{script_elements}",
        }
    }

    highlighter._handle_highlighjs_files_for(Note, anki_media_mock, anki_api_mock)

    anki_api_mock.update_note_type_templates.assert_not_called()


def test_handle_highlighjs_script_adds_scripts_to_multiple_templates(
    anki_api_mock, anki_media_mock
):
    note_type = Note
    anki_media_mock.exists.return_value = True
    anki_api_mock.fetch_note_type_templates.return_value = {
        "Card 1": {
            "Front": "{{Front}}",
            "Back": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}",
        },
        "Different": {"Front": "{{Front}} {{Front}}", "Back": "{{Back}}"},
    }
    script_elements = (
        f'<script src="_hl.pack.js"></script>\n{highlighter.AUTOSTART_SCRIPT}'
    )
    expected = {
        "Card 1": {
            "Front": "{{Front}}" + f"\n{script_elements}",
            "Back": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}"
            + f"\n{script_elements}",
        },
        "Different": {
            "Front": "{{Front}} {{Front}}" + f"\n{script_elements}",
            "Back": "{{Back}}" + f"\n{script_elements}",
        },
    }

    highlighter._handle_highlighjs_files_for(note_type, anki_media_mock, anki_api_mock)

    anki_api_mock.update_note_type_templates.assert_called_once_with(
        note_type, expected
    )



import re
from pathlib import Path
from typing import List, Optional, Union

from ..helpers import parse_str_to_bool
from ..mistune_plugins.mathjax3 import BLOCK_MATH
from .config import Config
from .notes.basic_note import BasicNote
from .notes.cloze_note import ClozeNote
from .notes.note import Note


class Parser:
    """Class for getting notes and various information about them from the text file"""

    _section_regex = re.compile(
        r"^---\n" r"(.+?)" r"^---$",  # start of section  # contents  # end of section
        re.MULTILINE | re.DOTALL,
    )
    _deck_name_regex = re.compile(r"(?<=^Deck:)(.*?)$", re.MULTILINE)
    _tags_regex = re.compile(r"(?<=^Tags:)(.*?)$", re.MULTILINE)
    _all_notes_regex = re.compile(
        r"(?:^<!--ID:\S+-->\n)?"  # optional ID
        r"^\d+\.[\s\S]*?"  # card contents (optionally contains answer)
        r"(?=<!--ID:\S+-->|^\d+\.|\Z|^---$)",  # card ends before ID, start of next card, end of string, end of section
        re.MULTILINE,
    )
    _basic_note_regex = re.compile(
        r"(?:^<!--ID:\S+-->\n)?"
        r"^\d+\.[\s\S]+?"
        r"(?:^>.*?(?:\n|$))+",  # optional ID  # card question  # card answer
        re.MULTILINE,
    )
    _cloze_note_regex = re.compile(
        r"(?:^<!--ID:\S+-->\n)?"  # optional ID
        r"^\d+\.[\s\S]*?{[\s\S]*?}[\s\S]*?"  # contents that must have '{' and '}' symbols
        r"(?=<!--ID:\S+-->|^\d+\.|\Z|^---$)",  # card ends before ID, start of next card, end of string, end of section
        re.MULTILINE,
    )
    _id_regex = re.compile(r"^<!--ID:(\S+)-->$", re.MULTILINE)
    _question_regex = re.compile(
        r"^\d+\."
        r"([\s\S]+?)"
        r"(?=^>|\Z)",  # question start  # contents  # answer start or end of string
        re.MULTILINE,
    )
    _answer_regex = re.compile(r"(?:^>.*?(?:\n|$))+", re.MULTILINE)

    def __init__(
        self,
        config: Config,
        file_path: Union[str, Path],
        default_deck: str,
    ):
        self._config = config
        self._file_path = file_path
        self._default_deck = default_deck

    def collect_notes(self) -> List[Note]:
        """Get all notes from the file which path was passed to the Parser"""
        with open(self._file_path, mode="rt", encoding="utf-8") as f:
            file_string = f.read()

        question_sections = self._get_sections(file_string)

        notes = []
        for section in question_sections:
            notes.extend(self._get_notes_from_section(section))

        return notes

    def _get_notes_from_section(self, section: str) -> List[Note]:
        """Get all Notes from the section string"""
        tags = self._get_tags(section)
        deck_name = self._get_deck_name(section)

        note_strings = self.get_note_strings(section)

        # Create note objects
        notes: List[Note] = []
        for string in note_strings:
            anki_id = self.get_id(string)

            # we check in this order because is_cloze_note_str can match front/back note if it contains curly braces
            if self._is_basic_note_str(string):
                question = self.get_question(string)
                answer = self._get_cleaned_answer(string)
                if not question or not answer:
                    continue

                notes.append(
                    BasicNote(
                        front_md=question,
                        back_md=self._add_filename_to_text(answer),
                        tags=tags,
                        deck_name=deck_name,
                        anki_id=anki_id,
                    )
                )
            elif self._is_cloze_note_str(string):
                text = self.get_question(string)
                if not text:
                    continue

                notes.append(
                    ClozeNote(
                        text_md=self._add_filename_to_text(text),
                        tags=tags,
                        deck_name=deck_name,
                        anki_id=anki_id,
                    )
                )
        return notes

    # wrapper to add string to end of text
    def _add_filename_to_text(self, text: str) -> str:
        """Add filename to the end of the text"""
        path_string = (
            f"""<span style="font-size: 9pt;">File: {self._file_path}</span>"""
        )
        if not parse_str_to_bool(
            self._config.get_option_value("defaults", "add_filename")
        ):
            return text
        return f"{text}\n\n{path_string}"

    def _get_deck_name(self, section: str) -> str:
        """Get deck name specified for this section"""
        matches = re.findall(Parser._deck_name_regex, section)

        # If no deck name
        if not matches:
            if not self._default_deck:
                raise ValueError(f"couldn't find deck name in section:\n{section}")

            return self._default_deck

        if len(matches) > 1:
            raise ValueError(f"more than one deck name field in section:\n{section}")

        deck_name = matches[0].strip()
        if not deck_name:
            raise ValueError(f"empty deck name field in section:\n{section}")

        return deck_name

    @classmethod
    def get_note_strings(cls, section: str) -> List[str]:
        """Get all strings of notes from section"""
        return re.findall(cls._all_notes_regex, section)

    @classmethod
    def get_id(cls, text: str) -> Optional[int]:
        """Get note's ID from text. Returns None if id wasn't found or if it is incorrect."""
        id_match = re.search(cls._id_regex, text)
        if id_match:
            try:
                return int(id_match.group(1))
            except ValueError:
                return None

        return None

    @classmethod
    def get_question(cls, text: str) -> Optional[str]:
        """Get clean question string from text
        (without digit followed by period and trailing whitespace)"""
        question_match = re.search(cls._question_regex, text)
        if question_match:
            return question_match.group(1).strip()

        return None

    @classmethod
    def get_answer(cls, text: str) -> Optional[str]:
        """Get answer string from text"""
        answer_match = re.search(cls._answer_regex, text)
        if answer_match:
            return answer_match.group()
        return None

    @classmethod
    def _get_sections(cls, file_contents: str) -> List[str]:
        """Get all sections (groups of notes) from the file string"""
        return re.findall(cls._section_regex, file_contents)

    @classmethod
    def _get_tags(cls, section: str) -> List[str]:
        """Get tags specified for this section"""
        matches = re.findall(cls._tags_regex, section)
        if not matches:
            return []

        if len(matches) > 1:
            raise ValueError(f"more than one tag field in section:\n{section}")

        tags = matches[0].strip().split()
        return tags

    @classmethod
    def _is_basic_note_str(cls, string: str) -> bool:
        """Check if note string contains basic note type"""
        match = re.search(cls._basic_note_regex, string)
        if not match:
            return False
        return True

    @classmethod
    def _is_cloze_note_str(cls, string: str) -> bool:
        """Check if note string contains cloze note type. Matches basic note type if it contains curly braces"""
        match = re.search(cls._cloze_note_regex, string)
        if not match:
            return False
        return True

    @classmethod
    def _get_cloze_note_strings(cls, section: str) -> List[str]:
        """Get all strings from section with only question and an (optional) ID"""
        return re.findall(cls._cloze_note_regex, section)

    @classmethod
    def _get_cleaned_answer(cls, text: str) -> Optional[str]:
        """Get clean answer string from text (without '>' and trailing whitespace)"""
        # Remove '>' and first whitespace char after it (if there is any)
        answer = cls.get_answer(text)
        if not answer:
            return None

        lines = answer.splitlines()
        cleaned_lines = []
        # Remove '>' and whitespace after it
        for line in lines:
            if len(line) > 1 and line[1].isspace():
                cleaned_lines.append(line[2:].rstrip())
            else:
                cleaned_lines.append(line[1:].rstrip())

        cleaned_answer = "\n\n".join(cleaned_lines)

        def replace_newlines(s: re.Match) -> str:
            return re.sub("\n\n", "\n", s.group(0))

        # change newlines in code blocks
        cleaned_answer = re.sub(r"```[\s\S]*?```", replace_newlines, cleaned_answer)

        # change newlines in math blocks
        cleaned_answer = re.sub(BLOCK_MATH, replace_newlines, cleaned_answer)

        return cleaned_answer



from typing import Optional

from inka2.models.notes.note import Note


class AnkiApiError(Exception):
    def __init__(self, message: str, note: Optional[Note] = None):
        super().__init__(message)
        self.note = note


class HighlighterError(Exception):
    pass



from typing import Any, Callable, Dict, Iterable, List, Optional

import mistune  # type: ignore
from rich.table import Column, Table

from ..config import Config
from .note import Note


class BasicNote(Note):
    """Front/Back note type"""

    def __init__(
        self,
        front_md: str,
        back_md: str,
        tags: Iterable[str],
        deck_name: str,
        anki_id: Optional[int] = None,
    ):
        super().__init__(tags, deck_name, anki_id)
        self.raw_front_md = front_md
        self.raw_back_md = back_md
        self.updated_front_md = front_md  # With updated image links
        self.updated_back_md = back_md  # With updated image links
        self.front_html = ""
        self.back_html = ""

    @property
    def search_query(self) -> str:
        """Query to search for note in Anki"""
        return self.create_anki_search_query(self.front_html)

    def convert_fields_to_html(
        self, convert_func: Callable[[str, mistune.Markdown], str], md: mistune.Markdown
    ) -> None:
        """Convert note fields from markdown to html using provided function"""
        self.front_html = convert_func(self.updated_front_md, md)
        self.back_html = convert_func(self.updated_back_md, md)

    def update_fields_with(self, update_func: Callable[[str], str]) -> None:
        """Updates values of *updated* fields using provided function"""
        self.updated_front_md = update_func(self.updated_front_md)
        self.updated_back_md = update_func(self.updated_back_md)

    def get_raw_fields(self) -> List[str]:
        """Get list of all raw (as in file) fields of this note"""
        return [self.raw_front_md, self.raw_back_md]

    def get_raw_question_field(self) -> str:
        """Get value of raw (as in file) question field"""
        return self.raw_front_md

    def get_html_fields(self, cfg: Config) -> Dict[str, str]:
        """Return dictionary with Anki field names as keys and html strings as values"""
        return {
            cfg.get_option_value("anki", "front_field"): self.front_html,
            cfg.get_option_value("anki", "back_field"): self.back_html,
        }

    @staticmethod
    def get_anki_note_type(cfg: Config) -> str:
        """Get name of Anki note type"""
        return cfg.get_option_value("anki", "basic_type")

    def __eq__(self, other: Any) -> bool:
        if not super().__eq__(other):
            return False

        return (
            self.raw_front_md == other.raw_front_md
            and self.raw_back_md == other.raw_back_md
        )

    def __rich__(self) -> Table:
        """Table that is used to display info about note in case of error"""
        table = Table(
            Column("Field", justify="left", style="magenta"),
            Column("Value", justify="left", style="green"),
            title="Basic Note",
        )
        table.add_row("Front", self.raw_front_md, end_section=True)
        table.add_row("Back", self.raw_back_md, end_section=True)
        table.add_row("Tags", ", ".join(self.tags), end_section=True)
        table.add_row("Deck", self.deck_name, end_section=True)

        return table

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(front_md={self.raw_front_md!r}, back_md={self.raw_back_md!r}, "
            f"tags={self.tags!r}, deck_name={self.deck_name!r}, anki_id={self.anki_id!r})"
        )



import pytest

from inka2.models.notes.note import Note

create_search_anki_query_test_cases = {
    # backward slashes
    r"$\sqrt{5}$": r'"$\\sqrt{5}$"',
    # double quotes
    'text and "quote"': r'"text and \"quote\""',
    # underscores
    "I have some_long_name in here": r'"I have some\_long\_name in here"',
    # colons
    "I like: colons a:b:c, yeah": r'"I like\: colons a\:b\:c, yeah"',
    # doesn't escape parenthesis
    "We have (much) to say": r'"We have (much) to say"',
    # doesn't escape hyphen
    "The hyphen - is a punctuation mark": r'"The hyphen - is a punctuation mark"',
}


@pytest.mark.parametrize("text, expected", create_search_anki_query_test_cases.items())
def test_escapes_colon(text, expected):
    assert Note.create_anki_search_query(text) == expected



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Iterable, List, Optional

import mistune  # type: ignore
from rich.table import Table

from inka2.models.config import Config


class Note(ABC):
    """Base class for all other note types"""

    def __init__(
        self, tags: Iterable[str], deck_name: str, anki_id: Optional[int] = None
    ):
        self.tags = tags
        self.deck_name = deck_name
        self.anki_id = anki_id
        self.changed = False  # Card was marked as changed in Anki
        self.to_delete = False  # Card was marked to be deleted in Anki

    @property
    @abstractmethod
    def search_query(self) -> str:
        """Query to search for note in Anki"""
        pass

    @abstractmethod
    def convert_fields_to_html(
        self, convert_func: Callable[[str, mistune.Markdown], str], md: mistune.markdown
    ) -> None:
        """Convert note fields from markdown to html using provided function"""
        pass

    @abstractmethod
    def update_fields_with(self, update_func: Callable[[str], str]) -> None:
        """Updates values of *updated* fields using provided function"""
        pass

    @abstractmethod
    def get_raw_fields(self) -> List[str]:
        """Get list of all raw (as in file) fields of this note"""
        pass

    @abstractmethod
    def get_raw_question_field(self) -> str:
        """Get value of raw (as in file) question field"""
        pass

    @abstractmethod
    def get_html_fields(self, cfg: Config) -> Dict[str, str]:
        """Return dictionary with Anki field names as keys and html strings as values"""
        pass

    @staticmethod
    @abstractmethod
    def get_anki_note_type(cfg: Config) -> str:
        """Get name of Anki note type"""
        pass

    @staticmethod
    def create_anki_search_query(text: str) -> str:
        """Create Anki search query from the supplied text"""
        special_chars = ["\\", '"', "*", "_", ":"]

        search_query = text
        for char in special_chars:
            escaped_char = "\\" + char
            search_query = search_query.replace(char, escaped_char)

        return '"' + search_query + '"'

    @abstractmethod
    def __rich__(self) -> Table:
        """Table that is used to display info about note in case of error"""
        pass

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self.tags == other.tags and self.deck_name == other.deck_name



import pytest

test_cases = {
    ("1. Question?\n" "\n" "Answer\n"): [],
    ("Tags:\n" "1. Question?\n" "\n" "Answer\n"): [],
    ("Tags: yolo\n" "1. Question?\n" "\n" "Answer\n"): ["yolo"],
    ("Tags: yolo abc new1\n" "1. Question?\n" "\n" "Answer\n"): ["yolo", "abc", "new1"],
    ("1. Question?\n" "\n" "Answer\n" "\n" "Tags: yolo\n" "\n" "2. Q?\n" "\n" "A\n"): [
        "yolo"
    ],
    ("Some text; Tags: yolo abc new1\n" "1. Question?\n" "\n" "Answer\n"): [],
}


@pytest.mark.parametrize("section, expected", test_cases.items())
def test_get_tags(fake_parser, section, expected):
    tags = fake_parser._get_tags(section)

    assert tags == expected


def test_get_tags_when_section_with_multiple_tag_fields_raises_error(fake_parser):
    section = (
        "Tags: yolo abc new1\n" "1. Question?\n" "\n" "Tags: okay bro\n" "Answer\n"
    )

    with pytest.raises(ValueError):
        fake_parser._get_tags(section)



import pytest
from click.testing import CliRunner

from inka2.cli import ROOT_DIR, cli, get_notes_from_file

# Collection of manual test cases


@pytest.mark.integration
def test_hello():
    UNDER_TEST = f"{ROOT_DIR}/tests/resources/test_inka_data.md"
    runner = CliRunner()
    result = runner.invoke(cli, ["collect", "-u", UNDER_TEST])
    assert result.exit_code == 0
    print(result.output)
    # assert 'Hello, Test!' in result.output


def test_duplicate_question_string_should_fail():
    UNDER_TEST = f"{ROOT_DIR}/tests/resources/index_out_of_range.md"
    runner = CliRunner()
    result = runner.invoke(cli, ["collect", "-u", UNDER_TEST])
    assert result.exit_code == 1


@pytest.mark.integration
def test_get_notes_from_file():
    notes = get_notes_from_file(f"{ROOT_DIR}/./tests/resources/test_inka_data.md")  # noqa
    _ = None


@pytest.mark.integration
def test_create_notes_from_files():
    """ creates anki card in default location "User 1"
    make init
    """
    UNDER_TEST = f"{ROOT_DIR}/tests/resources/test_inka_data.md"
    runner = CliRunner()
    result = runner.invoke(cli, ["-v", "collect", UNDER_TEST])
    assert result.exit_code == 0
    print(result.output)
    _ = None



import re
from typing import Iterable, Iterator, Match, Pattern, Tuple, Union

import mistune  # type: ignore

from ..mistune_plugins.mathjax3 import BLOCK_MATH, INLINE_MATH
from .notes.basic_note import Note
from .notes.cloze_note import ClozeNote

INLINE_CODE_REGEX = re.compile(r"`[\S\s]+?`", re.MULTILINE)
BLOCK_CODE_REGEX = re.compile(r"```[\s\S]+?```", re.MULTILINE)
CLOZE_DELETION_REGEX = re.compile(
    r"{{c\d+::[\s\S]*?}}|" r"{c?\d+::[\s\S]*?}|" r"{[\s\S]*?}", re.MULTILINE
)
ANKI_CLOZE_REGEX = re.compile(r"{{c\d+::[\s\S]*?}}", re.MULTILINE)
EXPLICIT_SHORT_CLOZE_REGEX = re.compile(r"{c?(\d+)::([\s\S]*?)}", re.MULTILINE)
IMPLICIT_SHORT_CLOZE_REGEX = re.compile(r"{([\s\S]*?)}", re.MULTILINE)

INLINE_CODE_PLACEHOLDER = "INLINE_CODE_PLACEHOLDER"
BLOCK_CODE_PLACEHOLDER = "BLOCK_CODE_PLACEHOLDER"
INLINE_MATH_PLACEHOLDER = "INLINE_MATH_PLACEHOLDER"
BLOCK_MATH_PLACEHOLDER = "BLOCK_MATH_PLACEHOLDER"


def convert_notes_to_html(notes: Iterable[Note], md: mistune.Markdown):
    """Convert note fields to html"""
    for note in notes:
        note.convert_fields_to_html(_convert_md_to_html, md)


def convert_cloze_deletions_to_anki_format(cloze_notes: Iterable[ClozeNote]):
    """Convert short syntax for cloze deletions to anki syntax"""
    for note in cloze_notes:
        # We substitute code and math inline/blocks with placeholders before searching cloze deletions
        # Block code should be removed before inline code
        block_code_matches, redacted_text = _get_matches_and_updated_text(
            BLOCK_CODE_REGEX, BLOCK_CODE_PLACEHOLDER, note.updated_text_md
        )
        inline_code_matches, redacted_text = _get_matches_and_updated_text(
            INLINE_CODE_REGEX, INLINE_CODE_PLACEHOLDER, redacted_text
        )

        # Block math should be removed before inline math
        block_math_matches, redacted_text = _get_matches_and_updated_text(
            BLOCK_MATH, BLOCK_MATH_PLACEHOLDER, redacted_text
        )
        inline_math_matches, redacted_text = _get_matches_and_updated_text(
            INLINE_MATH, INLINE_MATH_PLACEHOLDER, redacted_text
        )

        cloze_strings = CLOZE_DELETION_REGEX.findall(redacted_text)
        redacted_text = _replace_short_cloze_syntax(redacted_text, cloze_strings)

        redacted_text = _replace_placeholders_with_content(
            block_code_matches, BLOCK_CODE_PLACEHOLDER, redacted_text
        )
        redacted_text = _replace_placeholders_with_content(
            inline_code_matches, INLINE_CODE_PLACEHOLDER, redacted_text
        )
        redacted_text = _replace_placeholders_with_content(
            block_math_matches, BLOCK_MATH_PLACEHOLDER, redacted_text
        )
        redacted_text = _replace_placeholders_with_content(
            inline_math_matches, INLINE_MATH_PLACEHOLDER, redacted_text
        )

        note.updated_text_md = redacted_text


def _convert_md_to_html(text: str, md: mistune.Markdown) -> str:
    # We delete '\n' before and after each html tag because Anki is rendering them as newlines
    return re.sub(r"\n?(<.+?>)\n?", lambda tag_match: tag_match.group(1), md(text))


def _get_matches_and_updated_text(
    regex_pattern: Union[str, Pattern], placeholder: str, text: str
) -> Tuple[Iterator[Match[str]], str]:
    """Search for matches with regex pattern and then replace same matches in text with placeholder"""
    matches = re.finditer(regex_pattern, text)
    new_text = re.sub(regex_pattern, placeholder, text)
    return matches, new_text


def _replace_placeholders_with_content(
    matches: Iterable[Match[str]], placeholder: str, text: str
) -> str:
    """Replace placeholder occurrences with text from regex matches"""
    new_text = text
    for match in matches:
        new_text = new_text.replace(placeholder, match.group(0), 1)
    return new_text


def _replace_short_cloze_syntax(text: str, cloze_strings: Iterable[str]) -> str:
    """Replace short implicit and explicit cloze syntax with Anki cloze syntax in text"""
    new_text = text
    for i, cloze in enumerate(cloze_strings, start=1):
        if re.match(ANKI_CLOZE_REGEX, cloze):
            continue

        explicit_match = re.match(EXPLICIT_SHORT_CLOZE_REGEX, cloze)
        if explicit_match:
            explicit_index = explicit_match.group(1)
            content = explicit_match.group(2)
            new_cloze = "{{" + f"c{explicit_index}::{content}" + "}}"
            new_text = new_text.replace(cloze, new_cloze, 1)
            continue

        implicit_match = re.match(IMPLICIT_SHORT_CLOZE_REGEX, cloze)
        if implicit_match:
            content = implicit_match.group(1)
            new_cloze = "{{" + f"c{i}::{content}" + "}}"
            new_text = new_text.replace(cloze, new_cloze, 1)

    return new_text



from typing import Optional

from rich.console import Console

from inka2.models.notes.note import Note

CONSOLE = Console()


def parse_str_to_bool(string_value: str) -> bool:
    return string_value.strip().lower() == "true"


def print_result(msg: str) -> None:
    CONSOLE.print(f"{msg}", style="green")


def print_action(msg: str) -> None:
    CONSOLE.print(f"[cyan1]::[/cyan1] {msg}", style="bold")


def print_error(message: str, pause: bool = False, note: Optional[Note] = None) -> None:
    if note:
        CONSOLE.print(note)
    CONSOLE.print(f"ERROR: {message}", style="red bold")

    if pause:
        CONSOLE.input("Press [italic]Enter[/italic] to continue...\n")


def print_sub_error(msg: str) -> None:
    CONSOLE.print(f"   {msg}", style="red")


def print_warning(message: str) -> None:
    CONSOLE.print(f"WARNING: {message}", style="yellow bold")


def print_sub_warning(msg: str) -> None:
    CONSOLE.print(f"   {msg}", style="yellow")


def print_step(msg: str) -> None:
    CONSOLE.print(f"[green]==>[/green] {msg}", style="bold")


def print_sub_step(msg: str) -> None:
    CONSOLE.print(f"  [cyan1]->[/cyan1] {msg}")



import pytest

from inka2.models.notes.basic_note import BasicNote
from inka2.models.notes.cloze_note import ClozeNote
from inka2.models.parser import Parser


def test_get_notes_from_section_with_filename(config_add_filename):
    parser = Parser(config_add_filename, "file_doesnt_exist.md", "")
    section = "Deck: Abraham\n" "\n" "1. Some question?\n" "\n" "> Answer"
    expected = BasicNote(
        front_md="Some question?",
        back_md='Answer\n\n<span style="font-size: 9pt;">File: file_doesnt_exist.md</span>',
        tags=[],
        deck_name="Abraham",
    )
    cards = parser._get_notes_from_section(section)

    assert cards[0] == expected


test_cases = {
    "Deck: Abraham\n": [],
    ("Deck: Abraham\n" "Tags: some tags here"): [],
    ("Deck: Abraham\n" "\n" "1. Some question?\n" "\n" "> Answer"): [
        BasicNote(
            front_md="Some question?", back_md="Answer", tags=[], deck_name="Abraham"
        )
    ],
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): [
        BasicNote(
            front_md="Some question?",
            back_md="Answer",
            tags=["one", "two-three"],
            deck_name="Abraham",
        )
    ],
    (
        "Deck: Abraham\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "2. Q\n"
        "\n"
        "> A"
    ): [
        BasicNote(
            front_md="Some question?", back_md="Answer", tags=[], deck_name="Abraham"
        ),
        BasicNote(front_md="Q", back_md="A", tags=[], deck_name="Abraham"),
    ],
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "2. Q\n"
        "\n"
        "> A"
    ): [
        BasicNote(
            front_md="Some question?",
            back_md="Answer",
            tags=["one", "two-three"],
            deck_name="Abraham",
        ),
        BasicNote(
            front_md="Q", back_md="A", tags=["one", "two-three"], deck_name="Abraham"
        ),
    ],
    # cloze note
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some {question} here?\n"
        "\n"
    ): [
        ClozeNote(
            text_md="Some {question} here?",
            tags=["one", "two-three"],
            deck_name="Abraham",
        )
    ],
    # cloze note with ID
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:123456-->\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!\n"
        "\n"
    ): [
        ClozeNote(
            text_md="Some {question} here?\n\nMore info on question.\n\nAnd even more!",
            tags=["one", "two-three"],
            deck_name="Abraham",
            anki_id=123456,
        )
    ],
    # multiple cloze notes
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "2. {1::another} here?\n"
        "\n"
    ): [
        ClozeNote(
            text_md="Some {question} here?\n\nMore info on question.\n\nAnd even more!",
            tags=["one", "two-three"],
            deck_name="Abraham",
        ),
        ClozeNote(
            text_md="{1::another} here?", tags=["one", "two-three"], deck_name="Abraham"
        ),
    ],
    # multiple cloze notes with IDs
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:1612509025074-->\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "<!--ID:1612509015034-->\n"
        "2. {1::another} here?\n"
        "\n"
    ): [
        ClozeNote(
            text_md="Some {question} here?\n\nMore info on question.\n\nAnd even more!",
            tags=["one", "two-three"],
            deck_name="Abraham",
            anki_id=1612509025074,
        ),
        ClozeNote(
            text_md="{1::another} here?",
            tags=["one", "two-three"],
            deck_name="Abraham",
            anki_id=1612509015034,
        ),
    ],
    # basic and cloze notes with IDs
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "2. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "3. {1::another} here?\n"
        "\n"
    ): [
        ClozeNote(
            text_md="Some {question} here?\n\nMore info on question.\n\nAnd even more!",
            tags=["one", "two-three"],
            deck_name="Abraham",
        ),
        BasicNote(
            front_md="Some question?",
            back_md="Answer",
            tags=["one", "two-three"],
            deck_name="Abraham",
        ),
        ClozeNote(
            text_md="{1::another} here?", tags=["one", "two-three"], deck_name="Abraham"
        ),
    ],
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "2. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "3. Question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!\n"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> And more to it\n"
        "\n"
        "4. Where am I?\n"
        "\n"
    ): [
        BasicNote(
            front_md="Some question?",
            back_md="Answer",
            tags=["one", "two-three"],
            deck_name="Abraham",
        ),
        ClozeNote(
            text_md="Some {question} here?\n\nMore info on question.\n\nAnd even more!",
            tags=["one", "two-three"],
            deck_name="Abraham",
        ),
        BasicNote(
            front_md="Question?\n\nMore info on question.\n\nAnd even more!",
            back_md="Answer\n\n\n\nAdditional info\n\nAnd more to it",
            tags=["one", "two-three"],
            deck_name="Abraham",
        ),
    ],
    # should not collect empty basic note
    ("Deck: Abraham\n" "\n" "1. \n" "\n" "> "): [],
    ("Deck: Abraham\n" "\n" "1. not empty \n" "\n" "> "): [],
    # should not collect empty basic note
    ("Deck: Abraham\n" "\n" "1. \n" "\n" "> not empty"): [],
    # should not collect empty question
    ("Deck: Abraham\n" "\n" "1. \n"): [],
}


@pytest.mark.parametrize("section, expected", test_cases.items())
def test_get_notes_from_section(fake_parser, section, expected):
    cards = fake_parser._get_notes_from_section(section)

    assert cards == expected



import logging
import os
import sys
from pathlib import Path
from subprocess import call
from typing import Iterable, List, Set

import click
import mistune  # type: ignore
import requests
from rich.prompt import Confirm, Prompt
from rich.traceback import install

from . import __version__
from .exceptions import AnkiApiError, HighlighterError
from .helpers import (
    CONSOLE,
    parse_str_to_bool,
    print_action,
    print_error,
    print_result,
    print_step,
    print_sub_error,
    print_sub_step,
    print_sub_warning,
    print_warning,
)
from .mistune_plugins.mathjax3 import plugin_mathjax3
from .models import converter, highlighter, img_handler
from .models.anki_api import AnkiApi
from .models.anki_media import AnkiMedia
from .models.config import Config
from .models.hasher import Hasher
from .models.notes.basic_note import BasicNote
from .models.notes.cloze_note import ClozeNote
from .models.notes.note import Note
from .models.parser import Parser
from .models.writer import Writer

ROOT_DIR = Path(__file__).parent.parent.parent.absolute()
log = logging.getLogger(__name__)

DEFAULT_ANKI_FOLDERS = {
    "win32": r"~\AppData\Roaming\Anki2",
    "linux": "~/.local/share/Anki2",
    "darwin": "~/Library/Application Support/Anki2",
}

FILE_EXTENSIONS = [".md", ".markdown"]
CONFIG_PATH = f"{os.path.dirname(__file__)}/config.ini"
HASHES_PATH = f"{os.path.dirname(__file__)}/hashes.json"

CONFIG = Config(CONFIG_PATH)
MD = mistune.create_markdown(
    plugins=["strikethrough", "footnotes", "table", plugin_mathjax3],
    escape=parse_str_to_bool(CONFIG.get_option_value("defaults", "escape_html")),
)

# pretty traceback
install(console=CONSOLE)


def get_notes_from_file(file_path: str) -> List[Note]:
    print_sub_step("Getting cards from the file...")
    # We need to change working directory because images in file can have relative path
    os.chdir(os.path.dirname(file_path))

    default_deck = CONFIG.get_option_value("defaults", "deck")
    notes = Parser(CONFIG, file_path, default_deck).collect_notes()
    # TODO: add filename
    notes_num = len(notes)
    if notes_num == 0:
        print_sub_step("Cards weren't found!")
    else:
        print_sub_step(f'Found {notes_num} {"cards" if notes_num > 1 else "card"}!')

    return notes


def create_notes_from_file(
    file_path: str,
    full_sync: bool,
    anki_api: AnkiApi,
    anki_media: AnkiMedia,
    hasher: Hasher,
    force: bool = False,
) -> None:
    """Get all notes from file and send them to Anki"""
    print_step(f'Collecting cards from "{file_path}"!')

    print_sub_step("Calculating hash...")
    curr_hash = hasher.calculate_hash(file_path)
    if not (full_sync or hasher.has_changed(file_path, curr_hash)):
        print_sub_step("The file hasn't changed since last sync!")
        return

    notes = get_notes_from_file(file_path)
    if not notes:
        print_sub_step("Updating information on file hash...")
        hasher.update_hash(file_path, curr_hash)
        return

    converter.convert_cloze_deletions_to_anki_format(
        note for note in notes if isinstance(note, ClozeNote)
    )
    writer = Writer(file_path, notes)
    writer.update_cloze_notes()

    print_sub_step("Handling images...")
    img_handler.handle_images_in(notes, anki_media, force=force)

    print_sub_step("Converting cards to the html...")
    converter.convert_notes_to_html(notes, MD)

    print_sub_step("Synchronizing changes and adding new cards...")
    for note in notes:
        try:
            if note.anki_id:
                anki_api.update_note(note)
                continue

            note.anki_id = anki_api.add_note(note)
        except AnkiApiError as e:
            print_error(str(e), note=e.note)

    print_sub_step("Adding IDs to cards in file...")
    writer.update_note_ids()

    print_sub_step("Updating information on file hash...")
    hasher.update_hash(file_path, hasher.calculate_hash(file_path))
    print_sub_step("Finished!")


def update_note_ids_in_file(file_path: str, anki_api: AnkiApi, anki_media: AnkiMedia):
    """Update IDs of notes in file by getting their IDs from Anki"""
    print_step(f'Updating IDs of cards in "{file_path}"!')

    notes = get_notes_from_file(file_path)
    if not notes:
        return

    converter.convert_cloze_deletions_to_anki_format(
        [note for note in notes if isinstance(note, ClozeNote)]
    )
    writer = Writer(file_path, notes)
    writer.update_cloze_notes()

    print_sub_step("Handling images...")
    img_handler.handle_images_in(notes, anki_media, copy_images=False)

    print_sub_step("Converting cards to the html...")
    converter.convert_notes_to_html(notes, MD)

    print_sub_step("Getting card IDs from Anki...")
    anki_api.update_note_ids(notes)

    print_sub_step("Adding IDs to cards in file...")
    writer.update_note_ids()
    print_sub_step("Finished!")


def sync(anki_api: AnkiApi) -> None:
    try:
        anki_api.sync()
    except AnkiApiError as e:
        print_sub_warning(f"{e}. Skipping...")


def get_file_paths_from_directory(dir_path: str, search_recursive: bool) -> Set[str]:
    os.chdir(dir_path)

    paths_to_files = set()
    sub_directories = []
    for item in os.scandir():
        # Get markdown files
        extension = os.path.splitext(item.name)[1]
        if item.is_file() and extension in FILE_EXTENSIONS:
            paths_to_files.add(os.path.realpath(item.path))
        # And all directories
        elif search_recursive and item.is_dir():
            sub_directories.append(os.path.realpath(item.path))

    # Get all markdown files from each sub directory
    for directory in sub_directories:
        paths_to_files |= get_file_paths_from_directory(directory, search_recursive)
        os.chdir(dir_path)

    return paths_to_files


def get_paths_to_files(paths: Set[str], recursive: bool) -> Set[str]:
    paths_to_files = set()
    initial_directory = os.getcwd()
    for path in paths:
        full_path = os.path.realpath(path)

        if os.path.isdir(full_path):
            paths_to_files |= get_file_paths_from_directory(full_path, recursive)
            os.chdir(initial_directory)
            continue

        paths_to_files.add(full_path)

    return paths_to_files


def handle_code_highlight(anki_api: AnkiApi, anki_media: AnkiMedia) -> None:
    for note_type in (BasicNote, ClozeNote):
        highlighter.add_code_highlight_to(
            note_type,  # type: ignore
            CONFIG.get_option_value("highlight", "style"),
            anki_api,
            anki_media,
        )


def handle_note_types(anki_api: AnkiApi) -> None:
    note_types = anki_api.fetch_note_types()
    curr_basic_type = CONFIG.get_option_value("anki", "basic_type")
    curr_front_field = CONFIG.get_option_value("anki", "front_field")
    curr_back_field = CONFIG.get_option_value("anki", "back_field")
    curr_cloze_type = CONFIG.get_option_value("anki", "cloze_type")
    curr_cloze_field = CONFIG.get_option_value("anki", "cloze_field")
    css = """.card {
  font-family: arial;
  font-size: 20px;
  text-align: left;
  color: black;
  background-color: white;
  line-height: 1.5em;
}

.cloze {
  font-weight: bold;
  color: blue;
}
.nightMode .cloze {
  color: lightblue;
}

code {
  background-color: #232831;
  color: #fa4545;
  border: 1px solid #030a1420;
  box-shadow: 0 0.1em #00000010;
  padding: 1px 2px;
  line-height: 1.4em;
  box-sizing: border-box;
  font-family: Menlo, Monaco, Consolas, "Courier New", monospace;
}
"""

    # Basic Note
    if (
        not (curr_basic_type and curr_front_field and curr_back_field)
        or curr_basic_type not in note_types
    ):
        basic_name = "Inka Basic"
        front_field = "Front"
        back_field = "Back"
        if basic_name not in note_types:
            anki_api.create_note_type(
                name=basic_name,
                fields=[front_field, back_field],
                css=css,
                card_templates=[
                    {
                        "Name": "Card 1",
                        "Front": "{{" + f"{front_field}" + "}}\n",
                        "Back": "{{FrontSide}}\n<hr id=answer>\n"
                        + "{{"
                        + f"{back_field}"
                        + "}}\n",
                    }
                ],
                is_cloze=False,
            )
        CONFIG.update_option_value("anki", "basic_type", basic_name)
        CONFIG.update_option_value("anki", "front_field", front_field)
        CONFIG.update_option_value("anki", "back_field", back_field)

    # Cloze Note
    if not (curr_cloze_type and curr_cloze_field) or curr_cloze_type not in note_types:
        cloze_name = "Inka Cloze"
        text_field = "Text"
        if cloze_name not in note_types:
            anki_api.create_note_type(
                name=cloze_name,
                fields=[text_field],
                css=css,
                card_templates=[
                    {
                        "Name": "Cloze",
                        "Front": "{{" + f"cloze:{text_field}" + "}}\n",
                        "Back": "{{" + f"cloze:{text_field}" + "}}\n",
                    }
                ],
                is_cloze=True,
            )
        CONFIG.update_option_value("anki", "cloze_type", cloze_name)
        CONFIG.update_option_value("anki", "cloze_field", text_field)


def check_note_types(anki_media: AnkiMedia, anki_api: AnkiApi) -> None:
    print_action("Checking note types...")
    try:
        handle_note_types(anki_api)
        handle_code_highlight(anki_api, anki_media)
    except KeyError:
        print_error(
            'your config file is corrupted. Please reset its state with the command "inka config --reset".'
        )
        sys.exit(1)
    except requests.exceptions.RequestException as e:
        print_error(f'something gone wrong while handling code highlight: "{e}"')
        sys.exit(1)
    except HighlighterError as e:
        print_warning(str(e))


def get_profile_from_user(profiles: List[str]) -> str:
    profile = Prompt.ask(
        "   Enter the name of profile you want to use", choices=profiles
    )

    if Confirm.ask("   Save profile as default?"):
        CONFIG.update_option_value("defaults", "profile", profile)

    return profile


def get_profile(prompt_user: bool, anki_api: AnkiApi) -> str:
    profiles = anki_api.get_profiles()

    if len(profiles) == 1:
        return profiles[0]

    if prompt_user:
        return get_profile_from_user(profiles)

    profile = CONFIG.get_option_value("defaults", "profile")
    if not profile:
        print_sub_warning("Default profile is not specified!")
        profile = get_profile_from_user(profiles)
    elif profile not in profiles:
        print_sub_error(f'Incorrect profile name in the config: "{profile}"!')
        profile = get_profile_from_user(profiles)

    return profile


def edit_config_file(ctx, param, value) -> None:
    if not value or ctx.resilient_parsing:
        return

    editor = os.getenv("EDITOR")
    if not editor:
        editor = "nano"

    with open(CONFIG_PATH, mode="at", encoding="utf-8") as f:
        call([editor, f.name])

    ctx.exit()


def reset_config_file(ctx, param, value) -> None:
    if not value or ctx.resilient_parsing:
        return

    CONFIG.reset()
    ctx.exit()


def list_config_options(ctx, param, value) -> None:
    if not value or ctx.resilient_parsing:
        return

    CONSOLE.print(CONFIG)
    ctx.exit()


@click.group()
@click.version_option(version=__version__)
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose logging")
def cli(verbose: bool) -> None:
    """Inka2 - command-line tool for adding flashcards from Markdown files to Anki.

    Documentation:\n
        https://github.com/sysid/inka2/wiki

    Github:\n
        https://github.com/sysid/inka2
    """
    log_fmt = "%(asctime)-15s %(levelname)-7s %(message)s"
    if verbose:
        logging.basicConfig(
            format=log_fmt, level=logging.DEBUG, datefmt="%m-%d %H:%M:%S"
        )
    else:
        logging.basicConfig(
            format=log_fmt, level=logging.INFO, datefmt="%m-%d %H:%M:%S"
        )
    logging.getLogger("requests").setLevel(logging.INFO)
    logging.getLogger("urllib3").setLevel(logging.INFO)
    log.debug("Logging enabled.")
    log.debug(CONFIG.get_formatted_options())


@cli.command()
@click.option(
    "-l",
    "--list",
    "list_options",
    is_flag=True,
    callback=list_config_options,
    is_eager=True,
    help="List all options set in the config file, along with their values.",
)
@click.option(
    "-r",
    "--reset",
    is_flag=True,
    callback=reset_config_file,
    is_eager=True,
    help="Reset config file to the default state.",
)
@click.option(
    "-e",
    "--edit",
    is_flag=True,
    callback=edit_config_file,
    is_eager=True,
    help="Open the config file using editor specified in EDITOR environment variable.",
)
@click.argument("name", required=True)
@click.argument("value", required=False)
def config(list_options: bool, reset: bool, edit: bool, name: str, value: str) -> None:
    """Get and set inka's configuration options.

    NAME - name of the config option. VALUE - new value for the config option.

    Examples:\n
        Get the current value of "defaults.profile" option:\n
            inka config defaults.profile

        Set a new value to the "defaults.profile" option:\n
            inka config defaults.profile "My Profile"
    """
    try:
        section, key = name.split(".")

        if not value:
            print_result(CONFIG.get_option_value(section, key))
            sys.exit(0)

        CONFIG.update_option_value(section, key, value)
    except (KeyError, ValueError):
        print_error(
            "incorrect name of a config option!\n"
            'To get list of all options use "--list" flag.'
        )
        sys.exit(1)


@cli.command()
@click.option(
    "-r", "--recursive", is_flag=True, help="Search for files in subdirectories."
)
@click.option(
    "-p",
    "--prompt",
    is_flag=True,
    help="Show prompt for profile name even if config contains default profile.",
)
@click.option(
    "-u",
    "--update-ids",
    "update_ids",
    is_flag=True,
    help="Update incorrect or absent IDs of cards from file(s) by looking for these cards in Anki.",
)
@click.option(
    "-i",
    "--ignore-errors",
    "ignore_errors",
    is_flag=True,
    help="The program won't pause in case of an error.",
)
@click.option(
    "-f",
    "--full-sync",
    "full_sync",
    is_flag=True,
    help="Collect cards from the file, even if the file hasn't changed since the last sync.",
)
@click.option(
    "--force",
    "force",
    is_flag=True,
    help="Copy images to Anki Media folder even if they already exist, i.e. overwrite.",
)
@click.argument(
    "paths", metavar="[PATH]...", nargs=-1, type=click.Path(exists=True), required=False
)
def collect(
    recursive: bool,
    prompt: bool,
    update_ids: bool,
    ignore_errors: bool,
    full_sync: bool,
    force: bool,
    paths: Iterable[str],
) -> None:
    """Get flashcards from files and add them to Anki. If flashcard already exists in Anki, the changes will be synced.

    If no PATH argument is passed, the program will use the path from config option 'defaults.folder'.

       [PATH]... - paths to files and/or directories

       Examples:\n
           Get cards from file:\n
               inka collect path/to/cards.md

           Get cards from all Markdown files in the directory:\n
               inka collect path/to/directory
    """
    # If no path specified as an argument, search for default path in config
    paths = set(paths)
    if not paths:
        default_path = os.path.expanduser(CONFIG.get_option_value("defaults", "folder"))
        if not default_path:
            print_error(
                "default folder is not specified in the config!\n"
                "You must pass the path to a file or a folder as an argument."
            )
            sys.exit(1)

        if not os.path.exists(default_path):
            print_error(f'default folder "{default_path}" does not exist!')
            sys.exit(1)

        paths.add(default_path)

    # Get path to Anki folder
    anki_path = CONFIG.get_option_value("anki", "path")
    if not anki_path:
        anki_path = os.path.expanduser(DEFAULT_ANKI_FOLDERS[sys.platform])
    log.debug(f"Anki path: {anki_path}")

    # Create instance of AnkiApi. Throws an error if path to anki is incorrect
    try:
        anki_api = AnkiApi(CONFIG, Path(anki_path))
    except AnkiApiError as e:
        print_error(str(e))
        sys.exit(1)

    # Get name of profile and select it in Anki
    print_action("Getting profile...")
    profile = get_profile(prompt, anki_api)
    log.debug(f"{profile=}")

    # Load collection of a profile
    print_action("Loading profile...")
    try:
        anki_api.load_collection(profile)
    except AnkiApiError as e:
        print_error(str(e), pause=False)
        sys.exit(1)

    # Sync changes with AnkiWeb
    print_action("Getting changes from AnkiWeb...")
    sync(anki_api)

    # Check correctness of note types
    anki_media = AnkiMedia(profile, anki_path)
    check_note_types(anki_media, anki_api)

    # Get paths to all files
    print_action("Searching Markdown files...")
    files = get_paths_to_files(paths, recursive)
    if not files:
        print_sub_warning("Markdown files not found!")
        sys.exit(0)
    log.debug(f"{files=}")

    # Perform action on notes from each file
    hasher = Hasher(HASHES_PATH)
    initial_directory = os.getcwd()
    for file in files:
        try:
            if update_ids:
                update_note_ids_in_file(file, anki_api, anki_media)
                continue

            create_notes_from_file(
                file, full_sync, anki_api, anki_media, hasher, force=force
            )
        except (
            OSError,
            ValueError,
            FileNotFoundError,
            FileExistsError,
        ) as e:
            print_error(
                f"{e}\nSkipping file! Consider re-running with --force.",
                pause=(not ignore_errors),
            )
        except AnkiApiError as e:
            print_error(f"{e}\nSkipping file!", pause=(not ignore_errors), note=e.note)
        finally:
            os.chdir(initial_directory)

    # Sync changes with AnkiWeb
    print_action("Synchronizing changes with AnkiWeb...")
    sync(anki_api)

    # Close collection to save changes
    anki_api.close()
    print_action("Everything is done!")



import pytest

test_cases = {
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:1235523-->\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): 1235523,
    ("<!--ID:215321-->\n" "1. Some question?\n" "\n" "> Answer"): 215321,
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): None,
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:-->\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): None,
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:123a-->\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): None,
    ("<!--ID:1612579125074-->\n" "32. Some {question?}\n" "\n"): 1612579125074,
}


@pytest.mark.parametrize("text, expected", test_cases.items())
def test_get_id(fake_parser, text, expected):
    assert fake_parser.get_id(text) == expected



import pytest

test_cases = {
    # empty
    "": [],
    # one basic note
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): ["1. Some question?\n" "\n" "> Answer"],
    # basic note, only one newline between question and answer
    ("Deck: Abraham\n" "\n" "1. Some question?\n" "> Answer"): [
        "1. Some question?\n" "> Answer"
    ],
    # two basic notes
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "2. Q\n"
        "\n"
        "> A"
    ): [("1. Some question?\n" "\n" "> Answer\n" "\n"), ("2. Q\n" "\n" "> A")],
    # two basic notes, only one newline between end of first note and start of next
    (
        "Deck: Abraham\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "2. Q\n"
        "\n"
        "> A"
    ): [
        ("1. Some question?\n" "\n" "> Answer\n"),
        ("2. Q\n" "\n" "> A"),
    ],
    # basic note with multiline question
    (
        "Deck: Abraham\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer"
    ): [
        "1. Some question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer"
    ],
    # basic note with multiline answer
    (
        "Deck: Abraham\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> \n"
        "> And more to it"
    ): [
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> \n"
        "> And more to it"
    ],
    # basic note with multiline question and answer
    (
        "Deck: Abraham\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> \n"
        "> And more to it"
    ): [
        "1. Some question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> \n"
        "> And more to it"
    ],
    # note with only question before basic note
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "2. Another question\n"
        "\n"
        "> Answer"
    ): [("1. Some question?\n" "\n"), ("2. Another question\n" "\n" "> Answer")],
    # basic note with ID
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:123456-->\n"
        "1. Some question?\n"
        "\n"
        "> Answer"
    ): ["<!--ID:123456-->\n" "1. Some question?\n" "\n" "> Answer"],
    # two basic notes with ID
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:123456-->\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "<!--ID:581925-->\n"
        "2. Q\n"
        "\n"
        "> A"
    ): [
        ("<!--ID:123456-->\n" "1. Some question?\n" "\n" "> Answer\n" "\n"),
        ("<!--ID:581925-->\n" "2. Q\n" "\n" "> A"),
    ],
    # cloze note
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some {question} here?\n"
        "\n"
    ): [("1. Some {question} here?\n" "\n")],
    # cloze note with ID
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:123456-->\n"
        "1. Some {question} here?\n"
        "\n"
    ): [("<!--ID:123456-->\n" "1. Some {question} here?\n" "\n")],
    # multiline cloze note
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
    ): [
        (
            "1. Some {question} here?\n"
            "\n"
            "More info on question.\n"
            "\n"
            "And even more!"
            "\n"
        )
    ],
    # multiple cloze notes
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "2. {1::another} here?\n"
        "\n"
    ): [
        (
            "1. Some {question} here?\n"
            "\n"
            "More info on question.\n"
            "\n"
            "And even more!"
            "\n"
        ),
        ("2. {1::another} here?\n" "\n"),
    ],
    # multiple cloze notes with IDs
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "<!--ID:1612509025074-->\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "<!--ID:1612509015034-->\n"
        "2. {1::another} here?\n"
        "\n"
    ): [
        (
            "<!--ID:1612509025074-->\n"
            "1. Some {question} here?\n"
            "\n"
            "More info on question.\n"
            "\n"
            "And even more!"
            "\n"
        ),
        ("<!--ID:1612509015034-->\n" "2. {1::another} here?\n" "\n"),
    ],
    # basic and cloze notes with IDs
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "2. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "3. {1::another} here?\n"
        "\n"
    ): [
        (
            "1. Some {question} here?\n"
            "\n"
            "More info on question.\n"
            "\n"
            "And even more!"
            "\n"
        ),
        ("2. Some question?\n" "\n" "> Answer\n" "\n"),
        ("3. {1::another} here?\n" "\n"),
    ],
    (
        "Deck: Abraham\n"
        "\n"
        "Tags: one two-three\n"
        "\n"
        "1. Some question?\n"
        "\n"
        "> Answer\n"
        "\n"
        "2. Some {question} here?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "3. Question?\n"
        "\n"
        "More info on question.\n"
        "\n"
        "And even more!"
        "\n"
        "> Answer\n"
        "> \n"
        "> Additional info\n"
        "> \n"
        "> And more to it\n"
        "\n"
    ): [
        ("1. Some question?\n" "\n" "> Answer\n" "\n"),
        (
            "2. Some {question} here?\n"
            "\n"
            "More info on question.\n"
            "\n"
            "And even more!"
            "\n"
        ),
        (
            "3. Question?\n"
            "\n"
            "More info on question.\n"
            "\n"
            "And even more!"
            "\n"
            "> Answer\n"
            "> \n"
            "> Additional info\n"
            "> \n"
            "> And more to it\n"
            "\n"
        ),
    ],
}


@pytest.mark.parametrize("section, expected", test_cases.items())
def test_get_note_strings(fake_parser, section, expected):
    note_strings = fake_parser.get_note_strings(section)

    assert note_strings == expected



import hashlib
import json


class Hasher:
    def __init__(self, path: str):
        self._path = path
        try:
            with open(path, mode="rt", encoding="utf-8") as f:
                self._hashes = json.load(f)
        except FileNotFoundError:
            self._hashes = {}

    def update_hash(self, filepath: str, new_hash: str) -> None:
        """Update hash value for this filepath in the datafile"""
        self._hashes[filepath] = new_hash
        self._save()

    def has_changed(self, filepath: str, curr_hash: str) -> bool:
        """Check if the hash of this file changed. Returns True if file doesn't have previous hash value"""
        try:
            return self._hashes[filepath] != curr_hash
        except KeyError:
            return True

    def reset_hashes(self) -> None:
        """Remove all hashes from the file"""
        self._hashes = {}
        self._save()

    def _save(self) -> None:
        with open(self._path, mode="wt", encoding="utf-8") as f:
            json.dump(self._hashes, f)

    @staticmethod
    def calculate_hash(filepath: str) -> str:
        """Calculate MD5 hash for the file"""
        with open(filepath, mode="rt", encoding="utf-8") as f:
            encoded_text = f.read().encode("utf-8")
        return hashlib.md5(encoded_text).hexdigest()
